# Copied from
# https://github.com/lji72/inference/blob/93695ca637b1828532464f4648ba650cbfb59d65/others/cloud/single_stage_detector/tensorflow/utils_tf.py

import itertools
from math import sqrt

import numpy as np


# This function is from https://github.com/kuangliu/pytorch-ssd.
def calc_iou_tensor(box1, box2):
    """Calculation of IoU based on two boxes tensor,
    Reference to https://github.com/kuangliu/pytorch-ssd
    input:
        box1 (N, 4)
        box2 (M, 4)
    output:
        IoU (N, M)
    """
    N = box1.shape[0]
    M = box2.shape[0]

    be1 = np.expand_dims(box1, 1).repeat(M, axis=1)
    be2 = np.expand_dims(box2, 0).repeat(N, axis=0)
    lt = np.maximum(be1[:, :, :2], be2[:, :, :2])
    rb = np.minimum(be1[:, :, 2:], be2[:, :, 2:])

    delta = rb - lt
    delta[delta < 0] = 0
    intersect = delta[:, :, 0] * delta[:, :, 1]

    delta1 = be1[:, :, 2:] - be1[:, :, :2]
    area1 = delta1[:, :, 0] * delta1[:, :, 1]
    delta2 = be2[:, :, 2:] - be2[:, :, :2]
    area2 = delta2[:, :, 0] * delta2[:, :, 1]

    iou = intersect / (area1 + area2 - intersect)
    return iou


def softmax_cpu(x, dim=-1):
    x = np.exp(x)
    s = np.expand_dims(np.sum(x, axis=dim), dim)
    return x / s


def dboxes_R34_coco():
    figsize = [1200, 1200]
    feat_size = [[50, 50], [25, 25], [13, 13], [7, 7], [3, 3], [3, 3]]
    steps = [(int(figsize[0] / fs[0]), int(figsize[1] / fs[1])) for fs in feat_size]
    # use the scales here: https://github.com/amdegroot/ssd.pytorch/blob/master/data/config.py
    scales = [
        (int(s * figsize[0] / 300), int(s * figsize[1] / 300))
        for s in [21, 45, 99, 153, 207, 261, 315]
    ]
    aspect_ratios = [[2], [2, 3], [2, 3], [2, 3], [2], [2]]
    dboxes = DefaultBoxes(figsize, feat_size, steps, scales, aspect_ratios)
    return dboxes


# This function is from https://github.com/kuangliu/pytorch-ssd.
class Encoder(object):
    """
    Inspired by https://github.com/kuangliu/pytorch-ssd
    Transform between (bboxes, labels) <-> SSD output

    dboxes: default boxes in size 8732 x 4,
        encoder: input ltrb format, output xywh format
        decoder: input xywh format, output ltrb format

    decode:
        input  : bboxes_in (Tensor 8732 x 4), scores_in (Tensor 8732 x nitems)
        output : bboxes_out (Tensor nboxes x 4), labels_out (Tensor nboxes)
        criteria : IoU threshold of bboexes
        max_output : maximum number of output bboxes
    """

    def __init__(self, dboxes):
        self.dboxes = dboxes(order="ltrb")
        # self.dboxes_xywh = dboxes(order="xywh").unsqueeze(dim=0)
        self.dboxes_xywh = np.expand_dims(dboxes(order="xywh"), 0)
        self.nboxes = self.dboxes.shape[0]
        # print("# Bounding boxes: {}".format(self.nboxes))
        self.scale_xy = dboxes.scale_xy
        self.scale_wh = dboxes.scale_wh

    def scale_back_batch(self, bboxes_in, scores_in, device):
        """
        Do scale and transform from xywh to ltrb
        suppose input Nx4xnum_bbox Nxlabel_numxnum_bbox
        """

        bboxes_in = bboxes_in.transpose([0, 2, 1])
        scores_in = scores_in.transpose([0, 2, 1])

        bboxes_in[:, :, :2] = self.scale_xy * bboxes_in[:, :, :2]
        bboxes_in[:, :, 2:] = self.scale_wh * bboxes_in[:, :, 2:]

        bboxes_in[:, :, :2] = (
            bboxes_in[:, :, :2] * self.dboxes_xywh[:, :, 2:] + self.dboxes_xywh[:, :, :2]
        )
        bboxes_in[:, :, 2:] = np.exp(bboxes_in[:, :, 2:]) * self.dboxes_xywh[:, :, 2:]

        # Transform format to ltrb
        left, top, right, bottom = (
            bboxes_in[:, :, 0] - 0.5 * bboxes_in[:, :, 2],
            bboxes_in[:, :, 1] - 0.5 * bboxes_in[:, :, 3],
            bboxes_in[:, :, 0] + 0.5 * bboxes_in[:, :, 2],
            bboxes_in[:, :, 1] + 0.5 * bboxes_in[:, :, 3],
        )

        bboxes_in[:, :, 1] = left
        bboxes_in[:, :, 0] = top
        bboxes_in[:, :, 3] = right
        bboxes_in[:, :, 2] = bottom

        return bboxes_in, softmax_cpu(scores_in, dim=-1)

    def decode_batch(self, bboxes_in, scores_in, criteria=0.45, max_output=200, device=0):
        bboxes, probs = self.scale_back_batch(bboxes_in, scores_in, device)

        bbox_list = []
        label_list = []
        score_list = []

        # output format: List[List[box], List[label], List[score]]
        for bbox, prob in zip(bboxes, probs):
            bbox, label, score = self.decode_single(bbox, prob, criteria, max_output)
            bbox_list.append(bbox)
            label_list.append(label)
            score_list.append(score)

        return [bbox_list, label_list, score_list]

    # perform non-maximum suppression
    def decode_single(self, bboxes_in, scores_in, criteria, max_output, max_num=200):
        # Reference to https://github.com/amdegroot/ssd.pytorch

        bboxes_out = []
        scores_out = []
        labels_out = []

        for i, score in enumerate(np.split(scores_in, scores_in.shape[1], 1)):
            # skip background
            # print(score[score>0.90])
            if i == 0:
                continue
            score = score.squeeze(1)
            mask = score > 0.05

            bboxes, score = bboxes_in[mask, :], score[mask]
            if score.shape[0] == 0:
                continue

            score_idx_sorted = np.argsort(score, axis=0)

            # select max_output indices
            score_idx_sorted = score_idx_sorted[-max_num:]
            candidates = []

            while score_idx_sorted.size > 0:
                idx = score_idx_sorted[-1].item()
                bboxes_sorted = bboxes[score_idx_sorted, :]
                bboxes_idx = np.expand_dims(bboxes[idx, :], 0)
                iou_sorted = calc_iou_tensor(bboxes_sorted, bboxes_idx).squeeze()
                # we only need iou < criteria
                score_idx_sorted = score_idx_sorted[iou_sorted < criteria]
                candidates.append(idx)

            bboxes_out.append(bboxes[candidates, :])
            scores_out.append(score[candidates])
            labels_out.extend([i] * len(candidates))

        bboxes_out = np.concatenate(bboxes_out, axis=0)
        labels_out = np.array(labels_out, dtype=np.long)
        scores_out = np.concatenate(scores_out, axis=0)

        max_ids = np.argsort(scores_out, axis=0)
        max_ids = max_ids[-max_output:]
        return bboxes_out[max_ids, :], labels_out[max_ids], scores_out[max_ids]


class DefaultBoxes(object):
    def __init__(
        self, fig_size, feat_size, steps, scales, aspect_ratios, scale_xy=0.1, scale_wh=0.2
    ):
        self.feat_size = feat_size
        self.fig_size_w, self.fig_size_h = fig_size

        self.scale_xy_ = scale_xy
        self.scale_wh_ = scale_wh

        # According to https://github.com/weiliu89/caffe
        # Calculation method slightly different from paper
        self.steps_w = [st[0] for st in steps]
        self.steps_h = [st[1] for st in steps]
        self.scales = scales
        fkw = self.fig_size_w // np.array(self.steps_w)
        fkh = self.fig_size_h // np.array(self.steps_h)
        self.aspect_ratios = aspect_ratios

        self.default_boxes = []
        # size of feature and number of feature
        for idx, sfeat in enumerate(self.feat_size):
            sfeat_w, sfeat_h = sfeat
            sk1 = scales[idx][0] / self.fig_size_w
            sk2 = scales[idx + 1][1] / self.fig_size_h
            sk3 = sqrt(sk1 * sk2)
            all_sizes = [(sk1, sk1), (sk3, sk3)]
            for alpha in aspect_ratios[idx]:
                w, h = sk1 * sqrt(alpha), sk1 / sqrt(alpha)
                all_sizes.append((w, h))
                all_sizes.append((h, w))
            for w, h in all_sizes:
                for i, j in itertools.product(range(sfeat_w), range(sfeat_h)):
                    cx, cy = (j + 0.5) / fkh[idx], (i + 0.5) / fkw[idx]
                    self.default_boxes.append((cx, cy, w, h))
        self.dboxes = np.array(self.default_boxes)
        self.dboxes.clip(min=0, max=1, out=self.dboxes)
        # For IoU calculation
        self.dboxes_ltrb = self.dboxes.copy()
        self.dboxes_ltrb[:, 0] = self.dboxes[:, 0] - 0.5 * self.dboxes[:, 2]
        self.dboxes_ltrb[:, 1] = self.dboxes[:, 1] - 0.5 * self.dboxes[:, 3]
        self.dboxes_ltrb[:, 2] = self.dboxes[:, 0] + 0.5 * self.dboxes[:, 2]
        self.dboxes_ltrb[:, 3] = self.dboxes[:, 1] + 0.5 * self.dboxes[:, 3]

    @property
    def scale_xy(self):
        return self.scale_xy_

    @property
    def scale_wh(self):
        return self.scale_wh_

    def __call__(self, order="ltrb"):
        if order == "ltrb":
            return self.dboxes_ltrb
        if order == "xywh":
            return self.dboxes


encoder = Encoder(dboxes_R34_coco())



"""
implementation of imagenet dataset
"""

import logging
import os

logging.basicConfig(level=logging.INFO)
log = logging.getLogger("coco")

MobileNetSSD_CLASSES = [
    label.strip()
    for label in open(os.path.join(os.path.dirname(__file__), "coco-label-paper.txt")).readlines()
]

MobileNetSSD_Large_CLASSES = [
    label.strip()
    for label in open(
        os.path.join(os.path.dirname(__file__), "coco-label-ssdmobilelarge.txt")
    ).readlines()
]



from typing import Any, List

__all__ = [
    "ResNet50",
    "SSDMobileNet",
    "SSDResNet34",
    "YOLOv5l",
    "YOLOv5m",
    "EfficientNetB0",
    "EfficientNetV2s",
    "YOLOv7w6Pose",
]

_class_modules = {
    "ResNet50": ".resnet50",
    "SSDMobileNet": ".ssd_mobilenet",
    "SSDResNet34": ".ssd_resnet34",
    "YOLOv5l": ".yolov5",
    "YOLOv5m": ".yolov5",
    "EfficientNetB0": ".efficientnet_b0",
    "EfficientNetV2s": ".efficientnet_v2_s",
    "YOLOv7w6Pose": ".yolov7_w6_pose",
}


def __getattr__(name: str) -> Any:
    import importlib

    module = importlib.import_module(_class_modules.get(name, "." + name), __name__)
    cls = getattr(module, name)
    globals()[name] = cls  # so that __getattr__ won't be called again
    return cls


def __dir__() -> List[str]:
    return __all__



import numpy as np


class PythonPoseDecoder:
    def __init__(self, nc, anchors, stride, conf_thres) -> None:
        self.nc = nc  # number of classes
        self.no = nc + 5  # number of outputs per anchor
        self.nkpt = 17
        self.no_kpt = 3 * self.nkpt  # number of outputs per anchor for keypoints
        self.anchors = anchors
        self.nl = anchors.shape[0]  # number of detection layers
        self.na = anchors.shape[1]  # number of anchors
        self.grid = [None for _ in range(self.nl)]
        self.anchor_grid = [None for _ in range(self.nl)]
        self.kpt_grid_x = [None for _ in range(self.nl)]
        self.kpt_grid_y = [None for _ in range(self.nl)]
        self.stride = stride
        self.conf_thres = conf_thres  # conf_thres

    def _xywh2xyxy(self, x):
        # Convert nx4 boxes from [x, y, w, h] to [x1, y1, x2, y2]
        # where xy1=top-left, xy2=bottom-right
        y = np.empty_like(x)
        y[:, 0] = x[:, 0] - x[:, 2] / 2  # top left x
        y[:, 1] = x[:, 1] - x[:, 3] / 2  # top left y
        y[:, 2] = x[:, 0] + x[:, 2] / 2  # bottom right x
        y[:, 3] = x[:, 1] + x[:, 3] / 2  # bottom right y
        return y

    def _make_grid(self, nx=20, ny=20, i=0):
        xv, yv = np.meshgrid(np.arange(nx), np.arange(ny))
        grid = np.stack((xv, yv), axis=2).reshape((1, 1, ny, nx, 2)).astype(np.float32)
        anchor_grid = (
            (np.copy(self.anchors[i]) * self.stride[i]).reshape((1, self.na, 1, 1, 2))
            * np.ones((1, self.na, ny, nx, 2))
        ).astype(np.float32)

        kpt_grid_x = np.tile(grid[..., 0:1], (1, 1, 1, 1, self.nkpt))
        kpt_grid_y = np.tile(grid[..., 1:2], (1, 1, 1, 1, self.nkpt))

        return grid, anchor_grid, kpt_grid_x, kpt_grid_y

    def __call__(self, y_dets_kpts):
        z = []
        for i, y in enumerate(y_dets_kpts):
            y_det = y[..., :6]
            y_kpt = y[..., 6:]

            bs, _, ny, nx, _ = y_det.shape

            if self.grid[i] is None:
                (
                    self.grid[i],
                    self.anchor_grid[i],
                    self.kpt_grid_x[i],
                    self.kpt_grid_y[i],
                ) = self._make_grid(nx, ny, i)

            y_det = 1 / (1 + np.exp(-y_det))  # Sigmoid function
            y_det[..., 0:2] = (y_det[..., 0:2] * 2.0 - 0.5 + self.grid[i]) * self.stride[i]  # xy
            y_det[..., 2:4] = (y_det[..., 2:4] * 2) ** 2 * self.anchor_grid[i]  # wh

            y_kpt[..., 0::3] = (y_kpt[..., ::3] * 2.0 - 0.5 + self.kpt_grid_x[i]) * self.stride[
                i
            ]  # xy
            y_kpt[..., 1::3] = (y_kpt[..., 1::3] * 2.0 - 0.5 + self.kpt_grid_y[i]) * self.stride[
                i
            ]  # xy
            y_kpt[..., 2::3] = 1 / (1 + np.exp(-y_kpt[..., 2::3]))  # Sigmoid function

            y = np.concatenate((y_det, y_kpt), axis=-1)
            z.append(y.reshape(bs, -1, 57))

        out = []
        z = np.concatenate(z, axis=1)

        for x in z:
            x[..., 4] *= x[..., 5]
            conf = x[..., 4]
            x = x[conf > self.conf_thres]

            box = self._xywh2xyxy(x[:, :4])
            conf = x[..., 4:5]
            kpt = x[..., 6:]

            x = np.concatenate((box, conf, np.zeros_like(conf), kpt), axis=1)
            out.append(x)

        return out



import pathlib
from typing import List

import numpy as np
from pydantic import BaseModel
import yaml

from ...types import PythonPostProcessor
from .pose_decoder import PythonPoseDecoder

# from furiosa.models.vision.yolov7_w6_pose import PoseEstimationResult


INPUT_SIZE = (640, 384)
OUTPUT_SHAPES = [
    (1, 18, 48, 80),
    (1, 153, 48, 80),
    (1, 18, 24, 40),
    (1, 153, 24, 40),
    (1, 18, 12, 20),
    (1, 153, 12, 20),
    (1, 18, 6, 10),
    (1, 153, 6, 10),
]
with open(pathlib.Path(__file__).parent / "cfg.yaml", "r") as config_file:
    ANCHORS = yaml.safe_load(config_file)["anchors"]


class Keypoint(BaseModel):
    x: float
    y: float
    confidence: float


class PoseEstimationResult(BaseModel):
    bounding_box: List[float]
    confidence: float

    nose: Keypoint
    left_eye: Keypoint
    right_eye: Keypoint
    left_ear: Keypoint
    right_ear: Keypoint
    left_shoulder: Keypoint
    right_shoulder: Keypoint
    left_elbow: Keypoint
    right_elbow: Keypoint
    left_wrist: Keypoint
    right_wrist: Keypoint
    left_hip: Keypoint
    right_hip: Keypoint
    left_knee: Keypoint
    right_knee: Keypoint
    left_ankle: Keypoint
    right_ankle: Keypoint


def _nms(box_scores: np.ndarray, iou_threshold: float = 0.45) -> np.ndarray:
    output = []
    for box_scores in box_scores:
        scores = box_scores[:, 4]
        boxes = box_scores[:, :4]
        picked = []
        indexes = np.argsort(scores)[::-1]
        while len(indexes) > 0:
            current = indexes[0]
            picked.append(current.item())
            if len(indexes) == 1:
                break
            current_box = boxes[current, :]
            indexes = indexes[1:]
            rest_boxes = boxes[indexes, :]
            iou = _box_iou(rest_boxes, np.expand_dims(current_box, axis=0))
            indexes = indexes[iou <= iou_threshold]
        output.append(box_scores[picked, :])
    return output


def _box_area(left_top: np.ndarray, right_bottom: np.ndarray):
    """Compute the areas of rectangles given two corners."""
    # https://github.com/mlcommons/inference/blob/de6497f9d64b85668f2ab9c26c9e3889a7be257b/vision/classification_and_detection/python/models/utils.py#L89-L100
    width_height = np.clip(right_bottom - left_top, a_min=0.0, a_max=None)
    return width_height[..., 0] * width_height[..., 1]


def _box_iou(boxes1: np.ndarray, boxes2: np.ndarray, eps: float = 1e-5) -> np.ndarray:
    """Return intersection-over-union (Jaccard index) of boxes."""
    # https://github.com/mlcommons/inference/blob/de6497f9d64b85668f2ab9c26c9e3889a7be257b/vision/classification_and_detection/python/models/utils.py#L103-L119
    overlap_left_top = np.maximum(boxes1[..., :2], boxes2[..., :2])
    overlap_right_bottom = np.minimum(boxes1[..., 2:], boxes2[..., 2:])
    overlap_area = _box_area(overlap_left_top, overlap_right_bottom)
    area1 = _box_area(boxes1[..., :2], boxes1[..., 2:])
    area2 = _box_area(boxes2[..., :2], boxes2[..., 2:])
    return overlap_area / (area1 + area2 - overlap_area + eps)


class YOLOv7w6PosePostProcessor(PythonPostProcessor):
    def __init__(
        self,
        input_size=INPUT_SIZE,
        output_shapes=OUTPUT_SHAPES,
        anchors=ANCHORS,
        conf_thres=0.1,
        iou_thres=0.45,
    ) -> None:
        self.height = input_size[1]  # height only
        self.output_shapes = output_shapes
        self.anchors = np.float32(anchors)
        self.nkpt = 17
        self.no_kpt = 3 * self.nkpt  ## number of outputs per anchor for keypoints
        self.class_names = ["person"]

        self.conf_thres = conf_thres
        self.iou_thres = iou_thres

        self.stride = self._compute_stride()
        self.pose_decoder = PythonPoseDecoder(
            nc=len(self.class_names),
            anchors=self.anchors,
            stride=self.stride,
            conf_thres=self.conf_thres,
        )

    def _get_class_count(self):
        return len(self.class_names)

    def _get_output_feat_count(self):
        return self.anchors.shape[0]

    def _get_anchor_per_layer_count(self):
        return self.anchors.shape[1]

    def _reshape_output(self, i, y_det, y_kpt):
        shape = (
            y_det.shape[0],
            self._get_anchor_per_layer_count(),
            self._get_class_count() + 5 + self.no_kpt,
            y_det.shape[2],
            y_det.shape[3],
        )
        return np.concatenate((y_det, y_kpt), axis=1).reshape(shape).transpose(0, 1, 3, 4, 2)

    def __call__(self, feats_batched, contexts):
        assert len(feats_batched), "feats_batched must not be empty"

        boxes_batched = []
        batch_size = feats_batched[0].shape[0]
        assert batch_size == len(contexts)

        for i in range(batch_size):
            scale, (padw, padh) = contexts[i]["scale"], contexts[i]["pad"]
            if isinstance(scale, (tuple, list)):
                assert len(scale) == 2 and scale[0] == scale[1]
                scale = scale[0]
            feats = [f[i : i + 1] for f in feats_batched]

            y_dets, y_kpts = feats[0::2], feats[1::2]
            y_dets_kpts = [
                self._reshape_output(j, y_det, y_kpt)
                for j, (y_det, y_kpt) in enumerate(zip(y_dets, y_kpts))
            ]
            boxes_kpts_dec = self.pose_decoder(y_dets_kpts)
            boxes_kpts_dec = _nms(boxes_kpts_dec, self.iou_thres)[0]

            # rescale boxes
            boxes_kpts_dec[:, [0, 2]] = (1 / scale) * (boxes_kpts_dec[:, [0, 2]] - padw)
            boxes_kpts_dec[:, [1, 3]] = (1 / scale) * (boxes_kpts_dec[:, [1, 3]] - padh)

            boxes_kpts_dec[:, 6 + 0 : 6 + 3 * self.nkpt : 3] = (1 / scale) * (
                boxes_kpts_dec[:, 6 + 0 : 6 + 3 * self.nkpt : 3] - padw
            )
            boxes_kpts_dec[:, 6 + 1 : 6 + 3 * self.nkpt : 3] = (1 / scale) * (
                boxes_kpts_dec[:, 6 + 1 : 6 + 3 * self.nkpt : 3] - padh
            )

            # rescale keypoints

            boxes_batched.append(boxes_kpts_dec)

        return self._format_output(boxes_batched)

    def _format_output(self, boxes_batched):
        out_batched = []
        for dets in boxes_batched:
            box = dets[:, 0:4]
            conf = dets[:, 4]
            keypoints = dets[:, 6 : 6 + 3 * self.nkpt]

            out_batched.append(
                {
                    "box": box,
                    "conf": conf,
                    "pose": keypoints,
                }
            )
        final_output = []

        keypoints = [
            "nose",
            "left_eye",
            "right_eye",
            "left_ear",
            "right_ear",
            "left_shoulder",
            "right_shoulder",
            "left_elbow",
            "right_elbow",
            "left_wrist",
            "right_wrist",
            "left_hip",
            "right_hip",
            "left_knee",
            "right_knee",
            "left_ankle",
            "right_ankle",
        ]
        for item in out_batched:
            parsed = []
            for i in range(len(item["box"])):
                result = {
                    "bounding_box": item["box"][i].tolist(),
                    "confidence": item["conf"][i].tolist(),
                }
                for key, (x, y, conf) in zip(keypoints, zip(*[iter(item["pose"][i])] * 3)):
                    result[key] = {"x": x, "y": y, "confidence": conf}

                parsed.append(PoseEstimationResult(**result))
            final_output.append(parsed)
        return final_output

    def _compute_stride(self):
        img_h = self.height
        output_shapes = self.output_shapes[::2]
        feat_h = np.float32([shape[2] for shape in output_shapes])
        strides = img_h / feat_h
        return strides



from functools import reduce
import os
from time import perf_counter
from typing import Any, Callable, List, Optional, Sequence, Type

from tqdm import tqdm

from furiosa import device
from furiosa.common.thread import asynchronous
from furiosa.device import furiosa_native_device_sync as device_sync
from furiosa.runtime.sync import create_runner

from .. import vision
from ..types import Model


def normalize(text: str) -> str:
    """Returns only casefolded alphabets and numbers

    Args:
        text: string to be normalized

    Returns:
        Normalized text
    """
    text = "".join(filter(lambda c: c.isalnum(), text))
    return text.casefold()


def prettified_task_type(model: Type[Model]):
    """Returns pretty string for model's task type

    Args:
        model: Model class instance

    Returns:
        Prettified string for model's task type
    """
    task_type = model.task_type.name
    return " ".join(map(lambda x: x.capitalize(), task_type.split("_")))


def get_model_list(filter_func: Optional[Callable[..., bool]] = None) -> List[List[str]]:
    """Get list of available model names and descriptions

    Args:
        filter_func: Filter function for model class

    Returns:
        List of available model names and descriptions
    """
    filter_func = filter_func or (lambda _: True)
    model_list = []
    for model_name in vision.__all__:
        model_cls = getattr(vision, model_name)
        if not filter_func(model_cls):
            continue
        model = model_cls()
        postproc_map = model.postprocessor_map
        if not postproc_map:
            raise ValueError(f"No postprocessor map found for {model_name.capitalize()}")
        postprocs = ', '.join(map(lambda x: x.name.capitalize(), postproc_map.keys()))
        # Model name, description, task type, available post process implementations
        model_list.append([model_name, model.__doc__, prettified_task_type(model), postprocs])
    return model_list


def get_model(model_name: str) -> Optional[Type[Model]]:
    """Get model for given model name string

    Args:
        model_name: Model name

    Returns:
        A model which have name of the given argument
    """
    for name in vision.__all__:
        if normalize(name) == normalize(model_name):
            return getattr(vision, name)
    return None


def get_device_mode(device_str: str) -> device.DeviceMode:
    """Get device type for given device type string

    Args:
        device_type: Device type string

    Returns:
        Device type
    """
    device_config = device.DeviceConfig.from_str(device_str)
    device_list = device_sync.find_device_files(device_config)
    if not device_list:
        raise ValueError(f"No device found for {device_str}")
    assert reduce(
        lambda x, y: x == y, map(lambda x: x.mode(), device_list)
    ), "All devices must have same device type"
    return device_list[0].mode()


def device_file_to_pe_count(device_mode: device.DeviceMode) -> int:
    """Get number of PEs for given device mode

    Args:
        device_mode: DeviceMode

    Returns:
        Number of PEs
    """
    if device_mode == device.DeviceMode.Single:
        return 1
    elif device_mode == device.DeviceMode.Fusion:
        return 2
    else:
        raise ValueError(f"Unsupported device mode: {device_mode}")


def get_pe_count_from_device_str(device_str: Optional[str]) -> int:
    """Get number of PEs for given device string

    Args:
        device_str: Device string

    Returns:
        Number of PEs
    """
    device_str = device_str or os.environ.get(
        "FURIOSA_DEVICES", os.environ.get("NPU_DEVNAME", None)
    )
    if device_str:
        device_mode = get_device_mode(device_str)
        return device_file_to_pe_count(device_mode)
    else:
        return 2


def decorate_with_bar(string: str) -> str:
    """Decorate given string with bar

    Args:
        string: String to decorate

    Returns:
        Decorated string
    """

    bar = "----------------------------------------------------------------------"
    return "\n".join([bar, string, bar])


def time_with_proper_suffix(t: float, digits: int = 5) -> str:
    """Returns time with proper suffix

    Args:
        t: Time in seconds
        digits: Number of digits after decimal point

    Returns:
        Time with proper suffix
    """

    units = iter(["sec", "ms", "us", "ns"])
    while t < 1:
        t *= 1_000
        next(units)
    return f"{t:.{digits}f} {next(units)}"


def decorate_result(
    total_time: float, queries: int, header: str = "", digits: int = 5, newline: bool = True
) -> str:
    """Decorate benchmark result

    Args:
        total_time: Total elapsed time
        queries: Number of queries
        header: Header string
        digits: Number of digits after decimal point
        newline: Whether to add newline at the end

    Returns:
        Decorated benchmark result
    """

    result = []
    result.append(decorate_with_bar(header))
    result.append(f"Total elapsed time: {time_with_proper_suffix(total_time, digits)}")
    result.append(f"QPS: {queries / total_time:.{digits}f}")
    result.append(
        f"Avg. elapsed time / sample: {time_with_proper_suffix(total_time / queries, digits)}"
    )
    if newline:
        result.append("")
    return "\n".join(result)


def run_inferences(
    model_cls: Type[Model],
    input_paths: Sequence[str],
    postprocess: Optional[str],
    device_str: Optional[str],
):
    """Run inferences on given model

    Args:
        model_cls: Model class
        input_paths: Input paths
        postprocess: Postprocess implementation
        device_str: Device string
    """

    warning = """WARN: the benchmark results may depend on the number of input samples,
sizes of the images, and a machine where this benchmark is running."""
    num_pe = get_pe_count_from_device_str(device_str)
    model = model_cls(postprocessor_type=postprocess) if postprocess else model_cls()
    queries = len(input_paths)
    print(f"Run benchmark on {queries} input samples ...")

    print(decorate_with_bar(warning))
    with create_runner(model.model_source(num_pe=num_pe), device=device_str) as runner:
        model_inputs, model_outputs = [], []
        initial_time = perf_counter()
        for input_path in tqdm(input_paths, desc="Preprocess"):
            model_inputs.append(model.preprocess(input_path))
        after_preprocess = perf_counter()
        for contexted_model_input in tqdm(model_inputs, desc="Inference"):
            model_input, context = contexted_model_input
            model_outputs.append([runner.run(model_input), context])
        after_npu = perf_counter()
        for contexted_model_output in tqdm(model_outputs, desc="Postprocess"):
            model_output, context = contexted_model_output
            model.postprocess(model_output, context)
        all_done = perf_counter()

    print(
        decorate_result(all_done - initial_time, queries, "Preprocess -> Inference -> Postprocess")
    )
    print(decorate_result(all_done - after_preprocess, queries, "Inference -> Postprocess"))
    print(decorate_result(after_npu - after_preprocess, queries, "Inference", newline=False))


def serve_model(
    model_cls: Type[Model],
    postprocess: Optional[str],
    host: str,
    port: int,
    device_str: Optional[str],
) -> Any:
    """Serve model

    Args:
        model_cls: Model class
        postprocess: Postprocess implementation
        host: Host address
        port: Port number
        device_str: Device string
    """

    num_pe = get_pe_count_from_device_str(device_str)

    try:
        from fastapi import FastAPI, File, UploadFile

        from furiosa.serving import ServeAPI, ServeModel
    except ImportError:
        raise ImportError("Please install `furiosa-serving` to use this command")

    from tempfile import NamedTemporaryFile

    import numpy as np
    import uvicorn

    from furiosa.common.thread import synchronous

    serve = ServeAPI()
    app: FastAPI = serve.app
    model = model_cls(postprocessor_type=postprocess) if postprocess else model_cls()

    # ServeModel does not support in-memory model binary for now,
    # so we write model into temp file and pass its path
    model_file = NamedTemporaryFile()
    model_file.write(model.model_source(num_pe=num_pe))
    model_file_path = model_file.name

    serve_model: ServeModel = synchronous(serve.model("furiosart"))(
        model.name, location=model_file_path, npu_device=device_str
    )

    @serve_model.post("/infer")
    async def infer(image: UploadFile = File(...)):
        # Model Zoo's preprocesses do not consider in-memory image file for now
        # (note that it's different from in-memory tensor)
        # so we write in-memory image into temp file and pass its path
        image_file_path = NamedTemporaryFile()
        image_file_path.write(await image.read())

        tensors, ctx = await asynchronous(model.preprocess)(image_file_path.name)

        # Infer from ServeModel
        result: List[np.ndarray] = await serve_model.predict(tensors)

        return {"result": model.postprocess(result, ctx)}

    return uvicorn.run(app, host=host, port=port)



from itertools import product
from multiprocessing import Pool
import os
from pathlib import Path
import re
import shlex
import subprocess as sp
import tempfile
from typing import Tuple

import onnx
import yaml

from furiosa.quantizer import ModelEditor, TensorType, get_pure_input_names, quantize

COMPILER_CONFIG = {"lower_tabulated_dequantize": True}

compiler_info = sp.run(
    ['furiosa-compiler', '--version'], capture_output=True, text=True, check=True
).stdout
pattern = r"backend:\n- version: (.+?)\n- revision: (.+?)\b"
match = re.search(pattern, compiler_info)
if match:
    version = match.group(1)
    git_hash = match.group(2)
else:
    raise RuntimeError("Cannot parse furiosa-compiler version info")

base_path = Path(__file__).parent
print(f"furiosa-compiler {version} (rev. {git_hash})")
model_directories = [p for p in base_path.glob('*') if p.is_dir() and 'generated' not in p.name]
model_directories = list(enumerate(sorted(model_directories), start=1))
print(f"Found {len(model_directories)} ONNX model files:")
print('\n'.join(map(lambda x: f" {x[0]}. {x[1].name}", model_directories)))
generated_path = base_path / f"generated/{version}_{git_hash}"
generated_path.mkdir(parents=True, exist_ok=True)
print(f"Output directory: {generated_path}")


def quantize_and_compile_model(arg: Tuple[int, Path, int]):
    index, model_dir_path, num_pe = arg
    model_full_name = model_dir_path.name
    enf_path = generated_path / f"{model_full_name}_warboy_{num_pe}pe.enf"
    if enf_path.exists() and enf_path.is_file():
        print(f"  [{index}] {enf_path} already exists, so skipped", flush=True)
        return

    onnx_path = next(model_dir_path.glob('*_f32.onnx'))
    model_short_name = onnx_path.stem.removesuffix('_f32')
    calib_range_path = model_dir_path / f"{model_short_name}.calib_range.yaml"
    print(f"  [{index}] {model_short_name} starts from {onnx_path}", flush=True)

    # Load ONNX model
    onnx_model = onnx.load(onnx_path)

    # Quantize
    with open(calib_range_path) as f:
        calib_ranges = yaml.full_load(f)
    editor = ModelEditor(onnx_model)
    for input_name in get_pure_input_names(onnx_model):
        editor.convert_input_type(input_name, TensorType.UINT8)
    quantized_onnx = quantize(onnx_model, calib_ranges)
    with tempfile.NamedTemporaryFile() as quantized_onnx_file:
        onnx.save(quantized_onnx, quantized_onnx_file.name)
        print(f"  [{index}] {model_short_name} quantized", flush=True)

        # Acquire compiler config
        compiler_config = dict(COMPILER_CONFIG)
        additional_config = model_dir_path / "compiler_config.yaml"
        if additional_config.exists() and additional_config.is_file():
            print(f"   [{index}] {model_short_name} has extra compiler config: ", flush=True)
            with open(additional_config) as f:
                additional_compiler_config = yaml.full_load(f)
                print(str(additional_compiler_config), flush=True)
                compiler_config.update(additional_compiler_config)

        # Write compiler config to temp file
        fd, compiler_config_path = tempfile.mkstemp()
        with os.fdopen(fd, "w") as f:
            yaml.dump(compiler_config, f)

        # Compile and write to file
        target_npu = "warboy" if num_pe == 1 else "warboy-2pe"
        cmd = (
            f'furiosa-compiler --target-npu {target_npu} --output {enf_path} '
            f'{quantized_onnx_file.name}'
        )
        print(f"  [{index}] {model_short_name} compile start")
        sp.run(
            shlex.split(cmd),
            env={"NPU_COMPILER_CONFIG_PATH": compiler_config_path},
            check=True,
            stderr=sp.DEVNULL,
        )

        # Manually remove compiler config file (as we use mkstemp)
        Path(compiler_config_path).unlink()

        print(f"  [{index}] {model_short_name} compiled to {enf_path}", flush=True)


if __name__ == '__main__':
    lst = product(model_directories, [1, 2])  # [(idx, model_dir_path, num_pe), ...]
    lst = [(a, b, c) for (a, b), c in lst]  # Flatten list

    # Do the job parallelly
    with Pool() as pool:
        pool.map(quantize_and_compile_model, lst)



import os
from typing import Union

import cv2
import numpy as np
from numpy import ndarray


def center_crop(image: np.ndarray, cropped_height: int, cropped_width: int) -> np.ndarray:
    """Centrally crop `image` into cropped_width x cropped_height."""
    height, width, _ = image.shape
    top = int((height - cropped_height) / 2)
    bottom = int((height + cropped_height) / 2)
    left = int((width - cropped_width) / 2)
    right = int((width + cropped_width) / 2)
    image = image[top:bottom, left:right]
    return image


def resize_with_aspect_ratio(
    image: np.ndarray,
    scaled_height: int,
    scaled_width: int,
    percent: float,
    interpolation: int,
) -> np.ndarray:
    """Resize `image` so that it will be of scaled_width x scaled_height if it is scaled by
    `percent`.
    """
    height, width, _ = image.shape
    new_height = int(100.0 * scaled_height / percent)
    new_width = int(100.0 * scaled_width / percent)
    if height > width:
        new_height = int(new_height * height / width)
    else:
        new_width = int(new_width * width / height)
    image = cv2.resize(image, (new_width, new_height), interpolation=interpolation)
    return image


def read_image_opencv_if_needed(image: Union[str, os.PathLike, ndarray]):
    if isinstance(image, ndarray):
        return image
    elif isinstance(image, os.PathLike):
        path = image.__fspath__()  # imread only accepts str (opencv/opencv#15731)
    else:
        path = image
    image = cv2.imread(path)
    if image is None:
        raise FileNotFoundError(path)
    return image



from dataclasses import dataclass
from typing import Iterator, List, Sequence

import numpy as np


@dataclass
class CXcywhBoundingBox:
    center_x: float
    center_y: float
    width: float
    height: float

    def __iter__(self) -> Iterator[float]:
        return iter([self.center_x, self.center_y, self.width, self.height])


@dataclass
class XywhBoundingBox:
    x: float
    y: float
    width: float
    height: float

    def __iter__(self) -> Iterator[float]:
        return iter([self.x, self.y, self.width, self.height])


@dataclass
class LtrbBoundingBox:
    left: float
    top: float
    right: float
    bottom: float

    def __iter__(self) -> Iterator[float]:
        return iter([self.left, self.top, self.right, self.bottom])


def sigmoid(x: np.ndarray) -> np.ndarray:  # pylint: disable=invalid-name
    return 1 / (1 + np.exp(-x))


def calibration_ltrbbox(bbox: np.ndarray, width: float, height: float) -> np.ndarray:
    bbox[:, 0] *= width
    bbox[:, 1] *= height
    bbox[:, 2] *= width
    bbox[:, 3] *= height
    return bbox


def xyxytocxcywh(xyxy: LtrbBoundingBox) -> CXcywhBoundingBox:
    """Convert xyxy BoundingBox format to CXcywhBoundingBox format.

    Args:
        xyxy (LtrbBoundingBox): Left,Top,Right and Bottom bounding box format

    Returns:
        CXcywhBoundingBox: Center Point(x,y), width and height bounding box format
    """
    return CXcywhBoundingBox(
        center_x=(xyxy.left + xyxy.right) / 2,
        center_y=(xyxy.top + xyxy.bottom) / 2,
        width=xyxy.right - xyxy.left,
        height=xyxy.bottom + xyxy.top,
    )


def xyxytoxywh(xyxy: LtrbBoundingBox) -> XywhBoundingBox:
    """Convert xyxy BoundingBox format to XywhBoundingBox format.

    Args:
        xyxy (LtrbBoundingBox): Left,Top,Right and Bottom bounding box format

    Returns:
        CXcywhBoundingBox: LeftTop Point(x,y), width and height bounding box format
    """
    return XywhBoundingBox(
        x=xyxy.left,
        y=xyxy.top,
        width=xyxy.right - xyxy.left,
        height=xyxy.bottom + xyxy.top,
    )


@dataclass
class ObjectDetectionResult:
    boundingbox: LtrbBoundingBox
    score: float
    label: str
    index: int


def collate(data: Sequence[Sequence[np.array]], batch_axis=0) -> List[np.ndarray]:
    """This function converts a list of an numpy.array list into a batch type numpy array.
       The batch axis is specified according to the batch_axis.

    Args:
        data (List[List[np.array]]): a list of numpy.array list. The shape axis of each numpy.array
            must be 1.
        batch_axis (int, optional): batch axis. Defaults to 0.

    Returns:
        List[np.array]: a batch numpy array list.

    Examples:
        >>> arrays = [
            [np.random.randn(1, 3, 4), np.random.randn(1, 2, 5)],
            [np.random.randn(1, 3, 4), np.random.randn(1, 2, 5)],
            [np.random.randn(1, 3, 4), np.random.randn(1, 2, 5)],
        ]
        >>> batch_arrays = collate(arrays, batch_axis=0)
        >>> assert len(batch_arrays)==2
        >>> assert batch_arrays[0].shape==(3, 3, 4)
        >>> assert batch_arrays[1].shape==(3, 2, 5)
    """
    return [np.concatenate(x, axis=batch_axis) for x in zip(*data)]


def test_collate():
    arrays = [
        [np.random.randn(1, 3, 4), np.random.randn(1, 2, 5)],
        [np.random.randn(1, 3, 4), np.random.randn(1, 2, 5)],
        [np.random.randn(1, 3, 4), np.random.randn(1, 2, 5)],
    ]
    batch_arrays = collate(arrays, batch_axis=0)
    assert len(batch_arrays) == 2
    assert batch_arrays[0].shape == (3, 3, 4)
    assert batch_arrays[1].shape == (3, 2, 5)



"""YOLOv5l Module

Attributes:
    CLASSES (List[str]): a list of class names
"""
import pathlib
from typing import List, Union

import numpy as np
import yaml

from ..._utils import validate_postprocessor_type
from ...types import Metadata, Platform, Publication
from .core import YOLOv5Base

with open(pathlib.Path(__file__).parent / "datasets/yolov5l/cfg.yaml", "r") as f:
    configuration = yaml.safe_load(f)
    _ANCHORS: np.array = np.float32(configuration["anchors"])  # aspect ratio: (width, height)
    CLASSES: List[str] = configuration["class_names"]

__all__ = ['CLASSES', 'YOLOv5l']


class YOLOv5l(YOLOv5Base):
    """YOLOv5 Large model"""

    classes: List[str] = CLASSES

    def __init__(self, *, postprocessor_type: Union[str, Platform] = Platform.RUST):
        postprocessor_type = Platform(postprocessor_type)
        validate_postprocessor_type(postprocessor_type, self.postprocessor_map.keys())
        super().__init__(
            name="YOLOv5Large",
            metadata=Metadata(
                description="YOLOv5 large model",
                publication=Publication(url="https://github.com/ultralytics/yolov5"),
            ),
            postprocessor=self.postprocessor_map[postprocessor_type](_ANCHORS, CLASSES),
        )

        self._artifact_name = "yolov5l"



import logging
from typing import Any, ClassVar, Dict, List, Sequence, Tuple, Type, Union

import cv2
import numpy as np
import numpy.typing as npt

from ..._utils import validate_postprocessor_type
from ...types import (
    Format,
    ImageClassificationModel,
    Metadata,
    Platform,
    PostProcessor,
    PreProcessor,
    Publication,
    PythonPostProcessor,
)
from ..common.datasets import imagenet1k
from ..preprocess import center_crop, read_image_opencv_if_needed, resize_with_aspect_ratio

CLASSES: List[str] = imagenet1k.ImageNet1k_CLASSES

logger = logging.getLogger(__name__)


class ResNet50PreProcessor(PreProcessor):
    @staticmethod
    def __call__(
        image: Union[str, npt.ArrayLike], with_scaling: bool = False
    ) -> Tuple[np.ndarray, None]:
        """Convert an input image to a model input tensor

        Args:
            image: A path of an image or
                an image loaded as a numpy array in BGR order.
            with_scaling: Whether to apply model-specific techniques that involve scaling the
                model's input and converting its data type to float32. Refer to the code to gain a
                precise understanding of the techniques used. Defaults to False.

        Returns:
            The first element of the tuple is a numpy array that meets the input requirements of the
                ResNet50 model. The second element of the tuple is unused in this model and has no
                value. To learn more information about the output numpy array, please refer to
                [Inputs](resnet50_v1.5.md#inputs).
        """
        # https://github.com/mlcommons/inference/blob/af7f5a0b856402b9f461002cfcad116736a8f8af/vision/classification_and_detection/python/main.py#L37-L39
        # https://github.com/mlcommons/inference/blob/af7f5a0b856402b9f461002cfcad116736a8f8af/vision/classification_and_detection/python/dataset.py#L168-L184
        image = read_image_opencv_if_needed(image)
        assert image.dtype == np.uint8
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        image = resize_with_aspect_ratio(
            image, 224, 224, percent=87.5, interpolation=cv2.INTER_AREA
        )
        image = center_crop(image, 224, 224)
        if with_scaling:
            image = np.asarray(image, dtype=np.float32)
            # https://github.com/mlcommons/inference/blob/af7f5a0b856402b9f461002cfcad116736a8f8af/vision/classification_and_detection/python/dataset.py#L178
            image -= np.array([123.68, 116.78, 103.94], dtype=np.float32)
        image = image.transpose([2, 0, 1])
        return image[np.newaxis, ...], None


class ResNet50PostProcessor(PythonPostProcessor):
    def __call__(self, model_outputs: Sequence[npt.ArrayLike], contexts: Any = None) -> str:
        """Convert the outputs of a model to a label string, such as car and cat.

        Args:
            model_outputs: the outputs of the model.
                Please learn more about the output of model,
                please refer to [Outputs](resnet50_v1.5.md#outputs).

        Returns:
            str: A classified label, e.g., "tabby, tabby cat".
        """
        return CLASSES[int(model_outputs[0]) - 1]


class ResNet50(ImageClassificationModel):
    """MLCommons ResNet50 model"""

    postprocessor_map: ClassVar[Dict[Platform, Type[PostProcessor]]] = {
        Platform.PYTHON: ResNet50PostProcessor,
    }

    def __init__(self, *, postprocessor_type: Union[str, Platform] = Platform.PYTHON):
        postprocessor_type = Platform(postprocessor_type)
        validate_postprocessor_type(postprocessor_type, self.postprocessor_map.keys())
        super().__init__(
            name="ResNet50",
            format=Format.ONNX,
            family="ResNet",
            version="v1.5",
            metadata=Metadata(
                description="ResNet50 v1.5 int8 ImageNet-1K",
                publication=Publication(url="https://arxiv.org/abs/1512.03385.pdf"),
            ),
            preprocessor=ResNet50PreProcessor(),
            postprocessor=self.postprocessor_map[postprocessor_type](),
        )

        self._artifact_name = "mlcommons_resnet50_v1.5"



import math
from pathlib import Path
from typing import Any, ClassVar, Dict, List, Sequence, Tuple, Type, Union

from PIL import Image
import numpy as np
import numpy.typing as npt

from ..._utils import validate_postprocessor_type
from ...types import (
    Format,
    ImageClassificationModel,
    Metadata,
    Platform,
    PostProcessor,
    PreProcessor,
    Publication,
    PythonPostProcessor,
)
from ..common.datasets import imagenet1k

CLASSES: List[str] = imagenet1k.ImageNet1k_CLASSES

IMAGENET_DEFAULT_MEAN = np.array((0.485, 0.456, 0.406), dtype=np.float32)[:, np.newaxis, np.newaxis]
IMAGENET_DEFAULT_STD = np.array((0.229, 0.224, 0.225), dtype=np.float32)[:, np.newaxis, np.newaxis]


# https://github.com/pytorch/vision/blob/7ba97196757229552cede54639be75e3a0a9959f/torchvision/transforms/functional.py#L386-L392
def resize(image: Image.Image, size: int, resample: Image.Resampling) -> Image.Image:
    """Resize `image` so that a smaller edge of the resulting image will be matched to `size`."""
    # https://github.com/pytorch/vision/blob/0b6b8d42e95a208df85033a46fc7984c74301b35/torchvision/transforms/functional.py#L463
    # https://github.com/pytorch/vision/blob/0b6b8d42e95a208df85033a46fc7984c74301b35/torchvision/transforms/functional.py#L363-L383
    width, height = image.size
    if width <= height:
        new_width = size
        new_height = int(size * height / width)
    else:
        new_width = int(size * width / height)
        new_height = size

    image = image.resize((new_width, new_height), resample=resample)
    return image


# https://github.com/pytorch/vision/blob/7ba97196757229552cede54639be75e3a0a9959f/torchvision/transforms/functional.py#L551
def center_crop(image: Image.Image, cropped_height: int, cropped_width: int) -> Image.Image:
    """Centrally crop `image` into cropped_width x cropped_height."""
    width, height = image.size

    # https://github.com/pytorch/vision/blob/7ba97196757229552cede54639be75e3a0a9959f/torchvision/transforms/functional.py#L587-L588
    top = int(round((height - cropped_height) / 2))
    left = int(round((width - cropped_width) / 2))

    # https://github.com/pytorch/vision/blob/7ba97196757229552cede54639be75e3a0a9959f/torchvision/transforms/functional_pil.py#L237
    image = image.crop((left, top, left + cropped_width, top + cropped_height))
    return image


class EfficientNetB0PreProcessor(PreProcessor):
    @staticmethod
    def __call__(
        image: Union[str, Path, npt.ArrayLike], with_scaling: bool = False
    ) -> Tuple[np.ndarray, None]:
        """Read and preprocess an image located at image_path.

        Args:
            image: A path of an image.
            with_scaling: Whether to apply model-specific techniques that involve scaling the
                model's input and converting its data type to float32. Refer to the code to gain a
                precise understanding of the techniques used. Defaults to False.

        Returns:
            The first element of the tuple is a numpy array that meets the input requirements of
                the model. The second element of the tuple is unused in this model and has no
                value. To learn more information about the output numpy array, please refer to
                [Inputs](efficientnet_b0.md#inputs).
        """

        if isinstance(image, (str, Path)):
            image = Image.open(image).convert("RGB")

        scale_size = int(math.floor(224 / 0.875))
        image = resize(image, scale_size, resample=Image.Resampling.BICUBIC)

        data = center_crop(image, 224, 224)
        data = np.transpose(data, axes=(2, 0, 1))
        assert data.dtype == np.uint8

        if with_scaling:
            data = np.asarray(data, dtype=np.float32)
            data /= 255

            data -= IMAGENET_DEFAULT_MEAN
            data /= IMAGENET_DEFAULT_STD

        return data[np.newaxis, ...], None


class EfficientNetB0PostProcessor(PythonPostProcessor):
    def __call__(self, model_outputs: Sequence[npt.ArrayLike], contexts: Any = None) -> str:
        """Convert the outputs of a model to a label string, such as car and cat.

        Args:
            model_outputs: the outputs of the model.
                Please learn more about the output of model,
                please refer to [Outputs](efficientnet_b0.md#outputs).

        Returns:
            str: A classified label, e.g., "jigsaw puzzle".
        """

        return CLASSES[int(np.argsort(model_outputs[0], axis=1)[:, ::-1][0, 0])]


class EfficientNetB0(ImageClassificationModel):
    """EfficientNet B0 model"""

    postprocessor_map: ClassVar[Dict[Platform, Type[PostProcessor]]] = {
        Platform.PYTHON: EfficientNetB0PostProcessor,
    }

    def __init__(self, *, postprocessor_type: Union[str, Platform] = Platform.PYTHON):
        postprocessor_type = Platform(postprocessor_type)
        validate_postprocessor_type(postprocessor_type, self.postprocessor_map.keys())
        super().__init__(
            name="EfficientNetB0",
            format=Format.ONNX,
            family="EfficientNet",
            version="1.0.2",
            metadata=Metadata(
                description="EfficientNetB0 ImageNet-1K",
                publication=Publication(url="https://arxiv.org/abs/1905.11946"),
            ),
            preprocessor=EfficientNetB0PreProcessor(),
            postprocessor=self.postprocessor_map[postprocessor_type](),
        )

        self._artifact_name = "efficientnet_b0"



class FuriosaModelException(Exception):
    """General exception caused by Furiosa Runtime"""

    def __init__(self, message: str):
        self._message = message

    def __repr__(self):
        return '{}'.format(self._message)

    def __str__(self):
        return self.__repr__()


class ArtifactNotFound(FuriosaModelException):
    """Certain artifact file not found"""

    def __init__(self, artifact_file_name: str):
        super().__init__(f"{artifact_file_name} is required, but missing")


class NotFoundInDVCRemote(ArtifactNotFound):
    """Certain artifact file's DVC exists, but cannot fetch from remote"""

    def __init__(self, artifact_file_name: str, md5sum: str):
        super().__init__(f"{artifact_file_name} not exists in DVC remote\nmd5sum: {md5sum}")


class VersionInfoNotFound(FuriosaModelException):
    """Could not retrieve compiler version information"""

    def __init__(self):
        super().__init__(
            "Could not retrieve furiosa compiler information. Try: `pip install furiosa-sdk`"
        )


class ExtraPackageRequired(FuriosaModelException):
    """Needs extra packges to quantize and compile"""

    def __init__(self):
        super().__init__(
            "Needs extra packges to quantize and compile manually. Try: `pip install "
            "furiosa-models[full]`"
        )



import logging
import os
from pathlib import Path
import re
import subprocess
from typing import TYPE_CHECKING, Collection, Optional, Tuple, Union

if TYPE_CHECKING:
    from .types import Platform

import requests
import yaml

from . import errors

EXT_CALIB_YAML = "calib_range.yaml"
EXT_ONNX = "onnx"
DATA_DIRECTORY_BASE = Path(__file__).parent / "data"
CACHE_DIRECTORY_BASE = Path(
    os.getenv(
        "FURIOSA_MODELS_CACHE_HOME",
        os.path.join(os.getenv("XDG_CACHE_HOME", Path.home() / ".cache"), "furiosa/models"),
    )
)
DVC_PUBLIC_HTTP_ENDPOINT = (
    "https://furiosa-public-artifacts.s3-accelerate.amazonaws.com/furiosa-artifacts/"
)

module_logger = logging.getLogger(__name__)


def get_ir_version() -> str:
    try:
        compiler_info = subprocess.run(
            ['furiosa-compiler', '--version'], capture_output=True, text=True, check=True
        ).stdout
    except FileNotFoundError:
        raise RuntimeError("furiosa-compiler is not installed")
    pattern = r"backend:\n- version: (.+?)\n- revision: (.+?)\b"
    match = re.search(pattern, compiler_info)
    if match:
        version = match.group(1)
        git_hash = match.group(2)
        return f"{version}_{git_hash}"
    else:
        raise RuntimeError("Cannot parse furiosa-compiler version info")


def find_dvc_cache_directory(path: Path) -> Optional[Path]:
    if path is None or path == path.parent:
        return None
    if (path / ".dvc").is_dir():
        return path / ".dvc" / "cache"
    return find_dvc_cache_directory(path.parent)


def parse_dvc_file(file_path: Path) -> Tuple[str, str, int]:
    dvc_path = Path(f"{file_path}.dvc")
    if not dvc_path.exists():
        raise FileNotFoundError(f"{dvc_path} not found, are you in development mode?")
    info_dict = yaml.safe_load(open(f"{file_path}.dvc").read())["outs"][0]
    md5sum = info_dict["md5"]
    return md5sum[:2], md5sum[2:], info_dict["size"]


def get_from_url(path: str, uri: Path, is_legacy_path: bool = False) -> bytes:
    url = f"{DVC_PUBLIC_HTTP_ENDPOINT}{path}"
    module_logger.debug(f"Fetching from remote: {url}")
    with requests.get(url) as resp:
        if resp.status_code != 200:
            if not is_legacy_path:
                # New dvc now stores data into /files/md5
                return get_from_url(f"files/md5/{path}", uri, True)
            raise errors.NotFoundInDVCRemote(uri, path)
        data = resp.content
        caching_path = CACHE_DIRECTORY_BASE / get_ir_version() / (uri.name)
        module_logger.debug(f"caching to {caching_path}")
        caching_path.parent.mkdir(parents=True, exist_ok=True)
        with open(caching_path, mode="wb") as f:
            f.write(data)
        return data


class ArtifactResolver:
    def __init__(self, uri: Union[str, Path]):
        self.uri = Path(uri)
        # Note: DVC_REPO is to locate local DVC directory not remote git repository
        self.dvc_cache_path = os.environ.get("DVC_REPO", find_dvc_cache_directory(Path.cwd()))
        if self.dvc_cache_path is not None:
            self.dvc_cache_path = Path(self.dvc_cache_path)
            if self.dvc_cache_path.is_symlink():
                self.dvc_cache_path = self.dvc_cache_path.readlink()
            module_logger.debug(f"Found DVC cache directory: {self.dvc_cache_path}")

    def _read(self, directory: str, filename: str) -> bytes:
        # Try to find local cached file
        local_cache_path = CACHE_DIRECTORY_BASE / get_ir_version() / (self.uri.name)
        if local_cache_path.exists():
            module_logger.debug(f"Local cache exists: {local_cache_path}")
            with open(local_cache_path, mode="rb") as f:
                return f.read()

        # Try to find real file along with DVC file (no DVC)
        if Path(self.uri).exists():
            module_logger.debug(f"Local file exists: {self.uri}")
            with open(self.uri, mode="rb") as f:
                return f.read()

        module_logger.debug(f"{self.uri} not exists, resolving DVC")
        if self.dvc_cache_path is not None:
            cached: Path = self.dvc_cache_path / directory / filename
            if cached.exists():
                module_logger.debug(f"DVC cache hit: {cached}")
                with open(cached, mode="rb") as f:
                    return f.read()
            else:
                module_logger.debug(f"DVC cache directory exists, but not having {self.uri}")

        # Fetching from remote
        return get_from_url(f"{directory}/{filename}", self.uri)

    def read(self) -> bytes:
        directory, filename, size = parse_dvc_file(self.uri)
        data = self._read(directory, filename)
        assert len(data) == size
        return data


def resolve_source(src_name: str, extension: str) -> bytes:
    full_path = next((DATA_DIRECTORY_BASE / src_name).glob(f'*.{extension}.dvc'))
    # Remove `.dvc` suffix
    full_path = full_path.with_suffix('')
    return ArtifactResolver(full_path).read()


def resolve_model_source(src_name: str, num_pe: int = 2) -> bytes:
    version_info = get_ir_version()
    if version_info is None:
        raise errors.VersionInfoNotFound()
    generated_path_base = DATA_DIRECTORY_BASE / f"generated/{version_info}"
    if not generated_path_base.exists():
        module_logger.warning("ENF does not exist. Trying to generate from source..")

        try:
            import onnx
            import yaml

            from furiosa.quantizer import ModelEditor, TensorType, get_pure_input_names, quantize
        except ImportError:
            raise errors.ExtraPackageRequired()
        module_logger.warning(f"Returning quantized ONNX for {src_name}")
        onnx_model = onnx.load_from_string(resolve_source(src_name, EXT_ONNX))
        calib_range = yaml.full_load(resolve_source(src_name, EXT_CALIB_YAML))
        editor = ModelEditor(onnx_model)
        for input_name in get_pure_input_names(onnx_model):
            editor.convert_input_type(input_name, TensorType.UINT8)
        return quantize(onnx_model, calib_range)
    file_name = f'{src_name}_warboy_{num_pe}pe.enf'
    return ArtifactResolver(generated_path_base / file_name).read()


def validate_postprocessor_type(
    postprocessor_type: "Platform", postprocessor_map: Collection["Platform"]
):
    if postprocessor_type not in postprocessor_map:
        raise ValueError(
            f"Not supported postprocessor type: {postprocessor_type}, "
            f"Available choices: {', '.join(postprocessor_map)}"
        )



import logging
from typing import Any, ClassVar, Dict, List, Sequence, Tuple, Type, Union

import cv2
import numpy
import numpy as np
import numpy.typing as npt

from . import anchor_generator  # type: ignore[import]
from .. import native
from ..._utils import validate_postprocessor_type
from ...types import (
    Format,
    Metadata,
    ObjectDetectionModel,
    Platform,
    PostProcessor,
    PreProcessor,
    Publication,
    PythonPostProcessor,
    RustPostProcessor,
)
from ..common.datasets import coco
from ..postprocess import LtrbBoundingBox, ObjectDetectionResult, calibration_ltrbbox, sigmoid
from ..preprocess import read_image_opencv_if_needed

NUM_OUTPUTS: int = 12
CLASSES = coco.MobileNetSSD_CLASSES
NUM_CLASSES = len(CLASSES) - 1  # remove background

logger = logging.getLogger(__name__)

# https://github.com/mlcommons/inference/blob/de6497f9d64b85668f2ab9c26c9e3889a7be257b/vision/classification_and_detection/python/models/ssd_mobilenet_v1.py#L155-L158
PRIORS = np.concatenate(
    [
        tensor.numpy()
        # pylint: disable=protected-access
        for tensor in anchor_generator.create_ssd_anchors()._generate(
            [(19, 19), (10, 10), (5, 5), (3, 3), (2, 2), (1, 1)]
        )
    ],
    axis=0,
)
PRIORS_Y1 = np.expand_dims(np.expand_dims(PRIORS[:, 0], axis=1), axis=0)
PRIORS_X1 = np.expand_dims(np.expand_dims(PRIORS[:, 1], axis=1), axis=0)
PRIORS_Y2 = np.expand_dims(np.expand_dims(PRIORS[:, 2], axis=1), axis=0)
PRIORS_X2 = np.expand_dims(np.expand_dims(PRIORS[:, 3], axis=1), axis=0)
PRIORS_WIDTHS = PRIORS_X2 - PRIORS_X1
PRIORS_HEIGHTS = PRIORS_Y2 - PRIORS_Y1
PRIORS_CENTER_X = PRIORS_X1 + 0.5 * PRIORS_WIDTHS
PRIORS_CENTER_Y = PRIORS_Y1 + 0.5 * PRIORS_HEIGHTS

del PRIORS_Y1, PRIORS_X1, PRIORS_Y2, PRIORS_X2, PRIORS


def _decode_boxes(rel_codes: np.ndarray) -> np.ndarray:
    # https://github.com/mlcommons/inference/blob/de6497f9d64b85668f2ab9c26c9e3889a7be257b/vision/classification_and_detection/python/models/utils.py#L149-L198

    # pylint: disable=invalid-name

    dy = np.expand_dims(rel_codes[:, :, 0], axis=2)
    dx = np.expand_dims(rel_codes[:, :, 1], axis=2)
    dh = np.expand_dims(rel_codes[:, :, 2], axis=2)
    dw = np.expand_dims(rel_codes[:, :, 3], axis=2)

    # https://github.com/mlcommons/inference/blob/de6497f9d64b85668f2ab9c26c9e3889a7be257b/vision/classification_and_detection/python/models/ssd_mobilenet_v1.py#L127
    # https://github.com/mlcommons/inference/blob/de6497f9d64b85668f2ab9c26c9e3889a7be257b/vision/classification_and_detection/python/models/ssd_mobilenet_v1.py#L166
    # https://github.com/mlcommons/inference/blob/de6497f9d64b85668f2ab9c26c9e3889a7be257b/vision/classification_and_detection/python/models/utils.py#L177-L180
    dx /= 10.0
    dy /= 10.0
    dw /= 5.0
    dh /= 5.0

    prediction_center_x = dx * SSDSmallConstant.PRIORS_WIDTHS + SSDSmallConstant.PRIORS_CENTER_X
    prediction_center_y = dy * SSDSmallConstant.PRIORS_HEIGHTS + SSDSmallConstant.PRIORS_CENTER_Y
    prediction_w = np.exp(dw) * SSDSmallConstant.PRIORS_WIDTHS
    prediction_h = np.exp(dh) * SSDSmallConstant.PRIORS_HEIGHTS

    prediction_boxes = np.concatenate(
        (
            prediction_center_x - 0.5 * prediction_w,
            prediction_center_y - 0.5 * prediction_h,
            prediction_center_x + 0.5 * prediction_w,
            prediction_center_y + 0.5 * prediction_h,
        ),
        axis=2,
    )
    return prediction_boxes


def _filter_results(
    scores: np.ndarray,
    boxes: np.ndarray,
    confidence_threshold: float,
    iou_threshold: float,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    # https://github.com/mlcommons/inference/blob/de6497f9d64b85668f2ab9c26c9e3889a7be257b/vision/classification_and_detection/python/models/ssd_mobilenet_v1.py#L197-L212
    selected_box_probs = []
    labels = []
    for class_index in range(1, scores.shape[1]):
        probs = scores[:, class_index]
        mask = probs > confidence_threshold
        probs = probs[mask]
        subset_boxes = boxes[mask, :]
        box_probs = np.concatenate((subset_boxes, probs.reshape(-1, 1)), axis=1)
        box_probs = _nms(box_probs, iou_threshold)
        selected_box_probs.append(box_probs)
        labels.append(np.full((box_probs.shape[0],), class_index, dtype=np.int64))
    selected_box_probs = np.concatenate(selected_box_probs)  # type: ignore[assignment]
    labels = np.concatenate(labels)  # type: ignore[assignment]
    return selected_box_probs[:, :4], labels, selected_box_probs[:, 4]


def _nms(box_scores: np.ndarray, iou_threshold: float) -> np.ndarray:
    # https://github.com/mlcommons/inference/blob/de6497f9d64b85668f2ab9c26c9e3889a7be257b/vision/classification_and_detection/python/models/utils.py#L122-L146
    scores = box_scores[:, -1]
    boxes = box_scores[:, :-1]
    picked = []
    indexes = np.argsort(scores)[::-1]
    while len(indexes) > 0:
        current = indexes[0]
        picked.append(current.item())
        if len(indexes) == 1:
            break
        current_box = boxes[current, :]
        indexes = indexes[1:]
        rest_boxes = boxes[indexes, :]
        iou = _box_iou(rest_boxes, np.expand_dims(current_box, axis=0))
        indexes = indexes[iou <= iou_threshold]
    return box_scores[picked, :]


def _box_iou(boxes1: np.ndarray, boxes2: np.ndarray, eps: float = 1e-5) -> np.ndarray:
    """Return intersection-over-union (Jaccard index) of boxes."""
    # https://github.com/mlcommons/inference/blob/de6497f9d64b85668f2ab9c26c9e3889a7be257b/vision/classification_and_detection/python/models/utils.py#L103-L119
    overlap_left_top = np.maximum(boxes1[..., :2], boxes2[..., :2])
    overlap_right_bottom = np.minimum(boxes1[..., 2:], boxes2[..., 2:])
    overlap_area = _box_area(overlap_left_top, overlap_right_bottom)
    area1 = _box_area(boxes1[..., :2], boxes1[..., 2:])
    area2 = _box_area(boxes2[..., :2], boxes2[..., 2:])
    return overlap_area / (area1 + area2 - overlap_area + eps)


def _box_area(left_top: np.ndarray, right_bottom: np.ndarray):
    """Compute the areas of rectangles given two corners."""
    # https://github.com/mlcommons/inference/blob/de6497f9d64b85668f2ab9c26c9e3889a7be257b/vision/classification_and_detection/python/models/utils.py#L89-L100
    width_height = np.clip(right_bottom - left_top, a_min=0.0, a_max=None)
    return width_height[..., 0] * width_height[..., 1]


class SSDSmallConstant(object):
    PRIORS_WIDTHS = PRIORS_WIDTHS
    PRIORS_HEIGHTS = PRIORS_HEIGHTS
    PRIORS_CENTER_X = PRIORS_CENTER_X
    PRIORS_CENTER_Y = PRIORS_CENTER_Y


class SSDMobileNetPreProcessor(PreProcessor):
    @staticmethod
    def __call__(
        images: Sequence[Union[str, np.ndarray]],
        with_scaling: bool = False,
    ) -> Tuple[npt.ArrayLike, List[Dict[str, Any]]]:
        """Preprocess input images to a batch of input tensors.

        Args:
            images: A list of paths of image files (e.g., JPEG, PNG)
                or a stacked image loaded as a numpy array in BGR order or gray order.
            with_scaling: Whether to apply model-specific techniques that involve scaling the
                model's input and converting its data type to float32. Refer to the code to gain a
                precise understanding of the techniques used. Defaults to False.

        Returns:
            The first element is 3-channel images of 300x300 in NCHW format,
                and the second element is a list of context about the original image metadata.
                This context data should be passed and utilized during post-processing.
                To learn more about the outputs of preprocess (i.e., model inputs),
                please refer to [Inputs](ssd_mobilenet.md#inputs).

        """
        # https://github.com/mlcommons/inference/blob/de6497f9d64b85668f2ab9c26c9e3889a7be257b/vision/classification_and_detection/python/main.py#L49-L51
        # https://github.com/mlcommons/inference/blob/de6497f9d64b85668f2ab9c26c9e3889a7be257b/vision/classification_and_detection/python/dataset.py#L242-L249
        batch_image = []
        batch_preproc_param = []
        if isinstance(images, str):
            images = [images]
        for image in images:
            image = read_image_opencv_if_needed(image)
            assert image.dtype == np.uint8

            if with_scaling:
                image = np.array(image, dtype=np.float32)
            if len(image.shape) < 3 or image.shape[2] != 3:
                image = cv2.cvtColor(image, cv2.COLOR_GRAY2RGB)
            else:
                image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
            width = image.shape[1]
            height = image.shape[0]
            image = cv2.resize(image, (300, 300), interpolation=cv2.INTER_LINEAR)
            if with_scaling:
                image -= 127
                image /= 127
            image = image.transpose([2, 0, 1])
            batch_image.append(image)
            batch_preproc_param.append({"width": width, "height": height})
        return np.stack(batch_image, axis=0), batch_preproc_param


class SSDMobileNetPythonPostProcessor(PythonPostProcessor):
    @staticmethod
    def __call__(
        model_outputs: Sequence[numpy.ndarray],
        contexts: Sequence[Dict[str, Any]],
        confidence_threshold: float = 0.3,
        iou_threshold: float = 0.6,
    ) -> List[List[ObjectDetectionResult]]:
        """Convert the outputs of this model to a list of bounding boxes, scores and labels

        Arguments:
            model_outputs: the outputs of the model. To learn more about the output of model,
                please refer to [Outputs](ssd_mobilenet.md#outputs).
            contexts: context coming from `preprocess()`

        Returns:
            Detected Bounding Box and its score and label represented as `ObjectDetectionResult`.
                To learn more about `ObjectDetectionResult`, 'Definition of ObjectDetectionResult'
                can be found below.

        Definitions of ObjectDetectionResult and LtrbBoundingBox:
            ::: furiosa.models.vision.postprocess.LtrbBoundingBox
                options:
                    show_source: true
            ::: furiosa.models.vision.postprocess.ObjectDetectionResult
                options:
                    show_source: true
        """
        assert (
            len(model_outputs) == NUM_OUTPUTS
        ), f"the number of model outputs must be {NUM_OUTPUTS}, but {len(model_outputs)}"
        batch_size = model_outputs[0].shape[0]
        # https://github.com/mlcommons/inference/blob/de6497f9d64b85668f2ab9c26c9e3889a7be257b/vision/classification_and_detection/python/models/ssd_mobilenet_v1.py#L94-L97
        class_logits = [
            output.transpose((0, 2, 3, 1)).reshape((batch_size, -1, NUM_CLASSES))
            for output in model_outputs[0::2]
        ]
        box_regression = [
            output.transpose((0, 2, 3, 1)).reshape((batch_size, -1, 4))
            for output in model_outputs[1::2]
        ]
        # https://github.com/mlcommons/inference/blob/de6497f9d64b85668f2ab9c26c9e3889a7be257b/vision/classification_and_detection/python/models/ssd_mobilenet_v1.py#L144-L166
        class_logits = np.concatenate(class_logits, axis=1)  # type: ignore[assignment]
        box_regression = np.concatenate(box_regression, axis=1)  # type: ignore[assignment]
        batch_scores = sigmoid(class_logits)  # type: ignore[arg-type]
        batch_boxes = _decode_boxes(box_regression)  # type: ignore[arg-type]

        # https://github.com/mlcommons/inference/blob/de6497f9d64b85668f2ab9c26c9e3889a7be257b/vision/classification_and_detection/python/models/ssd_mobilenet_v1.py#L178-L185
        batch_results = []
        for scores, boxes, preproc_params in zip(
            batch_scores, batch_boxes, contexts
        ):  # loop mini-batch
            width, height = preproc_params["width"], preproc_params["height"]
            boxes, labels, scores = _filter_results(
                scores,
                boxes,
                confidence_threshold=confidence_threshold,
                iou_threshold=iou_threshold,
            )
            cal_boxes = calibration_ltrbbox(boxes, width, height)
            predicted_result = []
            for box, label, score in zip(cal_boxes, labels, scores):
                bb_list = box.tolist()
                predicted_result.append(
                    ObjectDetectionResult(
                        index=label,
                        label=CLASSES[label],
                        score=score,
                        boundingbox=LtrbBoundingBox(
                            left=bb_list[0], top=bb_list[1], right=bb_list[2], bottom=bb_list[3]
                        ),
                    )
                )
            batch_results.append(predicted_result)
        return batch_results


class SSDMobileNetNativePostProcessor(RustPostProcessor):
    def __init__(self):
        self._native = native.ssd_mobilenet.RustPostProcessor()

    def __call__(
        self, model_outputs: Sequence[numpy.ndarray], contexts: Sequence[Dict[str, Any]]
    ) -> List[List[ObjectDetectionResult]]:
        """Convert the outputs of this model to a list of bounding boxes, scores and labels

        Arguments:
            model_outputs: the outputs of the model. To learn more about the output of model,
                please refer to [Outputs](ssd_mobilenet.md#outputs).
            contexts: context coming from `preprocess()`

        Returns:
            Detected Bounding Box and its score and label represented as `ObjectDetectionResult`.
                To learn more about `ObjectDetectionResult`, 'Definition of ObjectDetectionResult'
                can be found below.

        Definitions of ObjectDetectionResult and LtrbBoundingBox:
            ::: furiosa.models.vision.postprocess.LtrbBoundingBox
                options:
                    show_source: true
            ::: furiosa.models.vision.postprocess.ObjectDetectionResult
                options:
                    show_source: true
        """
        if len(model_outputs) == 0:
            return [[]]
        batch_size = model_outputs[0].shape[0]
        assert batch_size == len(
            contexts
        ), "batch size of model_outputs and len(contexts) must be same"

        all_results = []

        for i, context in enumerate(contexts):
            raw_results = self._native.eval(
                [s[i, :] for s in model_outputs[1::2]],
                [s[i, :] for s in model_outputs[0::2]],
            )

            results = []
            width = context['width']
            height = context['height']
            for value in raw_results:
                left = value.left * width
                right = value.right * width
                top = value.top * height
                bottom = value.bottom * height
                results.append(
                    ObjectDetectionResult(
                        index=value.class_id,
                        label=CLASSES[value.class_id],
                        score=value.score,
                        boundingbox=LtrbBoundingBox(left=left, top=top, right=right, bottom=bottom),
                    )
                )
            all_results.append(results)

        return all_results


class SSDMobileNet(ObjectDetectionModel):
    """MLCommons MobileNet v1 model"""

    postprocessor_map: ClassVar[Dict[Platform, Type[PostProcessor]]] = {
        Platform.PYTHON: SSDMobileNetPythonPostProcessor,
        Platform.RUST: SSDMobileNetNativePostProcessor,
    }

    def __init__(self, *, postprocessor_type: Union[str, Platform] = Platform.RUST):
        postprocessor_type = Platform(postprocessor_type)
        validate_postprocessor_type(postprocessor_type, self.postprocessor_map.keys())
        super().__init__(
            name="MLCommonsSSDMobileNet",
            format=Format.ONNX,
            family="MobileNetV1",
            version="v1.1",
            metadata=Metadata(
                description="SSD MobileNet model for MLCommons v1.1",
                publication=Publication(url="https://arxiv.org/abs/1704.04861.pdf"),
            ),
            preprocessor=SSDMobileNetPreProcessor(),
            postprocessor=self.postprocessor_map[postprocessor_type](),
        )

        self._artifact_name = "mlcommons_ssd_mobilenet_v1"



import itertools
import logging
from math import sqrt
from typing import Any, ClassVar, Dict, List, Sequence, Tuple, Type, Union

import cv2
import numpy
import numpy as np
import numpy.typing as npt
import torch
import torch.nn.functional as F

from .. import native
from ..._utils import validate_postprocessor_type
from ...types import (
    Format,
    Metadata,
    ObjectDetectionModel,
    Platform,
    PostProcessor,
    PreProcessor,
    Publication,
    PythonPostProcessor,
    RustPostProcessor,
)
from ..common.datasets import coco
from ..postprocess import LtrbBoundingBox, ObjectDetectionResult, calibration_ltrbbox
from ..preprocess import read_image_opencv_if_needed

NUM_OUTPUTS: int = 12
CLASSES = coco.MobileNetSSD_Large_CLASSES

logger = logging.getLogger(__name__)


##Inspired by https://github.com/kuangliu/pytorch-ssd
class Encoder(object):
    """
    Transform between (bboxes, labels) <-> SSD output

    dboxes: default boxes in size 8732 x 4,
        encoder: input ltrb format, output xywh format
        decoder: input xywh format, output ltrb format

    decode:
        input  : bboxes_in (Tensor 8732 x 4), scores_in (Tensor 8732 x nitems)
        output : bboxes_out (Tensor nboxes x 4), labels_out (Tensor nboxes)
        criteria : IoU threshold of bboexes
        max_output : maximum number of output bboxes
    """

    def __init__(self, dboxes):
        self.dboxes = dboxes(order="ltrb")
        self.dboxes_xywh = dboxes(order="xywh").unsqueeze(dim=0)
        self.nboxes = self.dboxes.size(0)
        # print("# Bounding boxes: {}".format(self.nboxes))
        self.scale_xy = torch.tensor(dboxes.scale_xy)
        self.scale_wh = torch.tensor(dboxes.scale_wh)

    def decode_batch(self, bboxes_in, scores_in, criteria=0.50, max_output=200):
        self.dboxes = self.dboxes.to(bboxes_in)
        self.dboxes_xywh = self.dboxes_xywh.to(bboxes_in)
        bboxes, probs = scale_back_batch(
            bboxes_in, scores_in, self.scale_xy, self.scale_wh, self.dboxes_xywh
        )
        boxes = []
        labels = []
        scores = []
        for bbox, prob in zip(bboxes.split(1, 0), probs.split(1, 0)):
            bbox = bbox.squeeze(0)
            prob = prob.squeeze(0)
            dbox, dlabel, dscore = self.decode_single(bbox, prob, criteria, max_output)
            boxes.append(dbox)
            labels.append(dlabel)
            scores.append(dscore)

        return [boxes, labels, scores]

    # perform non-maximum suppression
    def decode_single(self, bboxes_in, scores_in, criteria, max_output, max_num=200):
        # Reference to https://github.com/amdegroot/ssd.pytorch

        bboxes_out = []
        scores_out = []
        labels_out = []

        for i, score in enumerate(scores_in.split(1, 1)):
            # skip background
            if i == 0:
                continue

            score = score.squeeze(1)
            mask = score > 0.05

            bboxes, score = bboxes_in[mask, :], score[mask]
            if score.size(0) == 0:
                continue

            score_sorted, score_idx_sorted = score.sort(dim=0)

            # select max_output indices
            score_idx_sorted = score_idx_sorted[-max_num:]
            candidates = []

            while score_idx_sorted.numel() > 0:
                idx = score_idx_sorted[-1].item()
                bboxes_sorted = bboxes[score_idx_sorted, :]
                bboxes_idx = bboxes[idx, :].unsqueeze(dim=0)
                iou_sorted = calc_iou_tensor(bboxes_sorted, bboxes_idx).squeeze()
                # we only need iou < criteria
                score_idx_sorted = score_idx_sorted[iou_sorted < criteria]
                candidates.append(idx)

            bboxes_out.append(bboxes[candidates, :])
            scores_out.append(score[candidates])
            labels_out.extend([i] * len(candidates))

        bboxes_out, labels_out, scores_out = (
            torch.cat(bboxes_out, dim=0),
            torch.tensor(labels_out, dtype=torch.long),
            torch.cat(scores_out, dim=0),
        )

        _, max_ids = scores_out.sort(dim=0)
        max_ids = max_ids[-max_output:]
        return bboxes_out[max_ids, :], labels_out[max_ids], scores_out[max_ids]


@torch.jit.script
def calc_iou_tensor(box1, box2):
    """Calculation of IoU based on two boxes tensor,
    Reference to https://github.com/kuangliu/pytorch-ssd
    input:
        box1 (N, 4)
        box2 (M, 4)
    output:
        IoU (N, M)
    """
    N = box1.size(0)
    M = box2.size(0)

    be1 = box1.unsqueeze(1).expand(-1, M, -1)
    be2 = box2.unsqueeze(0).expand(N, -1, -1)

    # Left Top & Right Bottom
    lt = torch.max(be1[:, :, :2], be2[:, :, :2])
    rb = torch.min(be1[:, :, 2:], be2[:, :, 2:])
    delta = rb - lt
    delta.clone().masked_fill_(delta < 0, 0)
    intersect = delta[:, :, 0] * delta[:, :, 1]
    delta1 = be1[:, :, 2:] - be1[:, :, :2]
    area1 = delta1[:, :, 0] * delta1[:, :, 1]
    delta2 = be2[:, :, 2:] - be2[:, :, :2]
    area2 = delta2[:, :, 0] * delta2[:, :, 1]

    iou = intersect / (area1 + area2 - intersect)
    return iou


@torch.jit.script
def scale_back_batch(bboxes_in, scores_in, scale_xy, scale_wh, dboxes_xywh):
    """
    Do scale and transform from xywh to ltrb
    suppose input Nx4xnum_bbox Nxlabel_numxnum_bbox
    """
    bboxes_in = bboxes_in.permute(0, 2, 1)
    scores_in = scores_in.permute(0, 2, 1)

    bboxes_in[:, :, :2] = scale_xy * bboxes_in[:, :, :2]
    bboxes_in[:, :, 2:] = scale_wh * bboxes_in[:, :, 2:]
    bboxes_in[:, :, :2] = bboxes_in[:, :, :2] * dboxes_xywh[:, :, 2:] + dboxes_xywh[:, :, :2]
    bboxes_in[:, :, 2:] = bboxes_in[:, :, 2:].exp() * dboxes_xywh[:, :, 2:]
    # Transform format to ltrb
    left, top, right, bottom = (
        bboxes_in[:, :, 0] - 0.5 * bboxes_in[:, :, 2],
        bboxes_in[:, :, 1] - 0.5 * bboxes_in[:, :, 3],
        bboxes_in[:, :, 0] + 0.5 * bboxes_in[:, :, 2],
        bboxes_in[:, :, 1] + 0.5 * bboxes_in[:, :, 3],
    )
    bboxes_in[:, :, 0] = left
    bboxes_in[:, :, 1] = top
    bboxes_in[:, :, 2] = right
    bboxes_in[:, :, 3] = bottom
    return bboxes_in, F.softmax(scores_in, dim=-1)


def dboxes_R34_coco(figsize, strides):
    feat_size = [[50, 50], [25, 25], [13, 13], [7, 7], [3, 3], [3, 3]]
    steps = [(int(figsize[0] / fs[0]), int(figsize[1] / fs[1])) for fs in feat_size]
    scales = [
        (int(s * figsize[0] / 300), int(s * figsize[1] / 300))
        for s in [21, 45, 99, 153, 207, 261, 315]
    ]
    aspect_ratios = [[2], [2, 3], [2, 3], [2, 3], [2], [2]]
    dboxes = DefaultBoxes(figsize, feat_size, steps, scales, aspect_ratios)
    return dboxes


class DefaultBoxes(object):
    def __init__(
        self, fig_size, feat_size, steps, scales, aspect_ratios, scale_xy=0.1, scale_wh=0.2
    ):
        self.feat_size = feat_size
        self.fig_size_w, self.fig_size_h = fig_size

        self.scale_xy_ = scale_xy
        self.scale_wh_ = scale_wh

        # According to https://github.com/weiliu89/caffe
        # Calculation method slightly different from paper
        self.steps_w = [st[0] for st in steps]
        self.steps_h = [st[1] for st in steps]
        self.scales = scales
        fkw = self.fig_size_w // np.array(self.steps_w)
        fkh = self.fig_size_h // np.array(self.steps_h)
        self.aspect_ratios = aspect_ratios

        self.default_boxes = []
        # size of feature and number of feature
        for idx, sfeat in enumerate(self.feat_size):
            sfeat_w, sfeat_h = sfeat
            sk1 = scales[idx][0] / self.fig_size_w
            sk2 = scales[idx + 1][1] / self.fig_size_h
            sk3 = sqrt(sk1 * sk2)
            all_sizes = [(sk1, sk1), (sk3, sk3)]
            for alpha in aspect_ratios[idx]:
                w, h = sk1 * sqrt(alpha), sk1 / sqrt(alpha)
                all_sizes.append((w, h))
                all_sizes.append((h, w))
            for w, h in all_sizes:
                for i, j in itertools.product(range(sfeat_w), range(sfeat_h)):
                    cx, cy = (j + 0.5) / fkh[idx], (i + 0.5) / fkw[idx]
                    self.default_boxes.append((cx, cy, w, h))
        self.dboxes = torch.tensor(self.default_boxes)
        self.dboxes.clamp_(min=0, max=1)
        # For IoU calculation
        self.dboxes_ltrb = self.dboxes.clone()
        self.dboxes_ltrb[:, 0] = self.dboxes[:, 0] - 0.5 * self.dboxes[:, 2]
        self.dboxes_ltrb[:, 1] = self.dboxes[:, 1] - 0.5 * self.dboxes[:, 3]
        self.dboxes_ltrb[:, 2] = self.dboxes[:, 0] + 0.5 * self.dboxes[:, 2]
        self.dboxes_ltrb[:, 3] = self.dboxes[:, 1] + 0.5 * self.dboxes[:, 3]

    @property
    def scale_xy(self):
        return self.scale_xy_

    @property
    def scale_wh(self):
        return self.scale_wh_

    def __call__(self, order="ltrb"):
        if order == "ltrb":
            return self.dboxes_ltrb
        if order == "xywh":
            return self.dboxes


def _pick_best(detections, confidence_threshold):
    bboxes, classes, confidences = detections
    best = np.argwhere(confidences > confidence_threshold).squeeze(axis=1)
    return [pred[best].squeeze(axis=0) for pred in detections]


class SSDResNet34PreProcessor(PreProcessor):
    @staticmethod
    def __call__(
        images: Sequence[Union[str, np.ndarray]],
        with_scaling: bool = False,
    ) -> Tuple[npt.ArrayLike, List[Dict[str, Any]]]:
        """Preprocess input images to a batch of input tensors

        Args:
            images: A list of paths of image files (e.g., JPEG, PNG)
                or a stacked image loaded as a numpy array in BGR order or gray order.
            with_scaling: Whether to apply model-specific techniques that involve scaling the
                model's input and converting its data type to float32. Refer to the code to gain a
                precise understanding of the techniques used. Defaults to False.

        Returns:
            The first element is a list of 3-channel images of 1200x1200
                in NCHW format, and the second element is a list of context about the original image
                metadata. This context data should be passed and utilized during post-processing.
                To learn more about the outputs of preprocess (i.e., model inputs),
                please refer to [Inputs](ssd_resnet34.md#inputs).

        """
        # https://github.com/mlcommons/inference/blob/de6497f9d64b85668f2ab9c26c9e3889a7be257b/vision/classification_and_detection/python/main.py#L141
        # https://github.com/mlcommons/inference/blob/de6497f9d64b85668f2ab9c26c9e3889a7be257b/vision/classification_and_detection/python/main.py#L61-L63
        # https://github.com/mlcommons/inference/blob/de6497f9d64b85668f2ab9c26c9e3889a7be257b/vision/classification_and_detection/python/dataset.py#L252-L263
        batch_image = []
        batch_preproc_param = []
        if isinstance(images, str):
            images = [images]
        for image in images:
            image = read_image_opencv_if_needed(image)
            assert image.dtype == np.uint8

            if with_scaling:
                image = np.array(image, dtype=np.float32)
            if len(image.shape) < 3 or image.shape[2] != 3:
                image = cv2.cvtColor(image, cv2.COLOR_GRAY2RGB)
            else:
                image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
            width = image.shape[1]
            height = image.shape[0]
            image = cv2.resize(image, (1200, 1200), interpolation=cv2.INTER_LINEAR)
            if with_scaling:
                mean = np.array([0.485, 0.456, 0.406], dtype=np.float32)
                std = np.array([0.229, 0.224, 0.225], dtype=np.float32)
                image = image / 255.0 - mean
                image = image / std
            # https://github.com/mlcommons/inference/blob/de6497f9d64b85668f2ab9c26c9e3889a7be257b/vision/classification_and_detection/python/main.py#L143
            # https://github.com/mlcommons/inference/blob/de6497f9d64b85668f2ab9c26c9e3889a7be257b/vision/classification_and_detection/python/coco.py#L40
            # https://github.com/mlcommons/inference/blob/de6497f9d64b85668f2ab9c26c9e3889a7be257b/vision/classification_and_detection/python/coco.py#L91
            image = image.transpose([2, 0, 1])
            batch_image.append(image)
            batch_preproc_param.append({"width": width, "height": height})
        return np.stack(batch_image, axis=0), batch_preproc_param


class SSDResNet34PythonPostProcessor(PythonPostProcessor):
    @staticmethod
    def __call__(
        model_outputs: Sequence[np.ndarray],
        contexts: Sequence[Dict[str, Any]],
        confidence_threshold=0.05,
    ) -> List[List[ObjectDetectionResult]]:
        """Convert the outputs of this model to a list of bounding boxes, scores and labels

        Args:
            model_outputs: the outputs of the model. To learn more about the output of model,
                please refer to [Outputs](ssd_resnet34.md#outputs).
            contexts: context coming from `preprocess()`

        Returns:
            Detected Bounding Box and its score and label represented as
                `ObjectDetectionResult`.
                To learn more about `ObjectDetectionResult`, 'Definition of ObjectDetectionResult'
                can be found below.

        Definition of ObjectDetectionResult and LtrbBoundingBox:
            ::: furiosa.models.vision.postprocess.LtrbBoundingBox
                options:
                    show_source: true
            ::: furiosa.models.vision.postprocess.ObjectDetectionResult
                options:
                    show_source: true
        """
        if len(model_outputs) != NUM_OUTPUTS:
            raise Exception(f"output size must be {NUM_OUTPUTS}, but {len(model_outputs)}")
        classes, locations = model_outputs[:6], model_outputs[6:]

        # https://github.com/mlcommons/inference/blob/de6497f9d64b85668f2ab9c26c9e3889a7be257b/vision/classification_and_detection/python/models/ssd_r34.py#L317-L329
        classes = [np.reshape(cls, (cls.shape[0], 81, -1)) for cls in classes]
        locations = [np.reshape(loc, (loc.shape[0], 4, -1)) for loc in locations]
        classes = np.concatenate(classes, axis=2)
        locations = np.concatenate(locations, axis=2)

        # https://github.com/mlcommons/inference/blob/de6497f9d64b85668f2ab9c26c9e3889a7be257b/vision/classification_and_detection/python/models/ssd_r34.py#L251-L253
        # https://github.com/mlcommons/inference/blob/de6497f9d64b85668f2ab9c26c9e3889a7be257b/vision/classification_and_detection/python/models/ssd_r34.py#L350
        # len(det_boxes: List[np.array]) = N batch
        det_boxes, det_labels, det_scores = Encoder(
            dboxes_R34_coco((1200, 1200), (3, 3, 2, 2, 2, 2))
        ).decode_batch(torch.from_numpy(locations), torch.from_numpy(classes), 0.50, 200)

        # Pick the best boxes
        # https://pytorch.org/hub/nvidia_deeplearningexamples_ssd/
        # sometimes there are many boxes with localizaition and class probability distrituion.
        batch_results = []
        for boxes, labels, scores, preproc_params in zip(
            det_boxes, det_labels, det_scores, contexts
        ):
            width, height = preproc_params['width'], preproc_params['height']
            boxes, labels, scores = _pick_best(
                detections=(boxes, labels, scores),
                confidence_threshold=confidence_threshold,
            )
            cal_boxes = calibration_ltrbbox(boxes, width, height)
            predicted_result = []
            for box, label, score in zip(cal_boxes, labels, scores):
                bb_list = box.tolist()
                predicted_result.append(
                    ObjectDetectionResult(
                        index=int(label),
                        label=CLASSES[label],
                        score=score,
                        boundingbox=LtrbBoundingBox(
                            left=bb_list[0], top=bb_list[1], right=bb_list[2], bottom=bb_list[3]
                        ),
                    )
                )
            batch_results.append(predicted_result)
        return batch_results


class SSDResNet34NativePostProcessor(RustPostProcessor):
    def __init__(self):
        self._native = native.ssd_resnet34.RustPostProcessor()

    def __call__(
        self, model_outputs: Sequence[numpy.ndarray], contexts: Sequence[Dict[str, Any]]
    ) -> List[List[ObjectDetectionResult]]:
        """Convert the outputs of this model to a list of bounding boxes, scores and labels

        Args:
            model_outputs: the outputs of the model. To learn more about the output of model,
                please refer to [Outputs](ssd_resnet34.md#outputs).
            contexts: context coming from `preprocess()`

        Returns:
            Detected Bounding Box and its score and label represented as
                `ObjectDetectionResult`.
                To learn more about `ObjectDetectionResult`, 'Definition of ObjectDetectionResult'
                can be found below.

        Definition of ObjectDetectionResult and LtrbBoundingBox:
            ::: furiosa.models.vision.postprocess.LtrbBoundingBox
                options:
                    show_source: true
            ::: furiosa.models.vision.postprocess.ObjectDetectionResult
                options:
                    show_source: true
        """
        if len(model_outputs) == 0:
            return [[]]
        batch_size = model_outputs[0].shape[0]
        assert batch_size == len(
            contexts
        ), "batch size of model_outputs and len(contexts) must be same"

        all_results = []

        for i, context in enumerate(contexts):
            raw_results = self._native.eval(
                [s[i, :] for s in model_outputs[6:]],
                [s[i, :] for s in model_outputs[:6]],
            )

            results = []
            width = context['width']
            height = context['height']
            for value in raw_results:
                left = value.left * width
                right = value.right * width
                top = value.top * height
                bottom = value.bottom * height
                results.append(
                    ObjectDetectionResult(
                        index=value.class_id,
                        label=CLASSES[value.class_id],
                        score=value.score,
                        boundingbox=LtrbBoundingBox(left=left, top=top, right=right, bottom=bottom),
                    )
                )
            all_results.append(results)

        return all_results


class SSDResNet34(ObjectDetectionModel):
    """MLCommons SSD ResNet34 model"""

    postprocessor_map: ClassVar[Dict[Platform, Type[PostProcessor]]] = {
        Platform.PYTHON: SSDResNet34PythonPostProcessor,
        Platform.RUST: SSDResNet34NativePostProcessor,
    }

    def __init__(self, *, postprocessor_type: Union[str, Platform] = Platform.RUST):
        postprocessor_type = Platform(postprocessor_type)
        validate_postprocessor_type(postprocessor_type, self.postprocessor_map.keys())
        super().__init__(
            name="SSDResNet34",
            format=Format.ONNX,
            family="ResNet",
            version="v1.1",
            metadata=Metadata(
                description="SSD ResNet34 model for MLCommons v1.1",
                publication=Publication(
                    url="https://github.com/mlcommons/inference/tree/master/vision/classification_and_detection"
                    # noqa: E501
                ),
            ),
            preprocessor=SSDResNet34PreProcessor(),
            postprocessor=self.postprocessor_map[postprocessor_type](),
        )

        self._artifact_name = "mlcommons_ssd_resnet34"



import numpy as np
import torch

# The following functions were taken from
# https://github.com/tensorflow/models/tree/master/research/object_detection
# with minor modifications so that they use torch operations instead


def expanded_shape(orig_shape, start_dim, num_dims):
    s = (1,) * num_dims
    return orig_shape[:start_dim] + s + orig_shape[start_dim:]


def meshgrid(x, y):
    """Tiles the contents of x and y into a pair of grids.
    Multidimensional analog of numpy.meshgrid, giving the same behavior if x and y
    are vectors. Generally, this will give:
    xgrid(i1, ..., i_m, j_1, ..., j_n) = x(j_1, ..., j_n)
    ygrid(i1, ..., i_m, j_1, ..., j_n) = y(i_1, ..., i_m)
    Keep in mind that the order of the arguments and outputs is reverse relative
    to the order of the indices they go into, done for compatibility with numpy.
    The output tensors have the same shapes.  Specifically:
    xgrid.get_shape() = y.get_shape().concatenate(x.get_shape())
    ygrid.get_shape() = y.get_shape().concatenate(x.get_shape())
    Args:
    x: A tensor of arbitrary shape and rank. xgrid will contain these values
       varying in its last dimensions.
    y: A tensor of arbitrary shape and rank. ygrid will contain these values
       varying in its first dimensions.
    Returns:
    A tuple of tensors (xgrid, ygrid).
    """
    x = torch.as_tensor(x)
    y = torch.as_tensor(y)
    x_exp_shape = expanded_shape(x.shape, 0, y.dim())
    y_exp_shape = expanded_shape(y.shape, y.dim(), x.dim())

    xgrid = torch.reshape(x, x_exp_shape).repeat(*y_exp_shape)
    ygrid = torch.reshape(y, y_exp_shape).repeat(*x_exp_shape)
    new_shape = y.shape + x.shape
    xgrid = xgrid.reshape(new_shape)
    ygrid = ygrid.reshape(new_shape)

    return xgrid, ygrid


def tile_anchors(
    grid_height,
    grid_width,
    scales,
    aspect_ratios,
    base_anchor_size,
    anchor_stride,
    anchor_offset,
):
    """Create a tiled set of anchors strided along a grid in image space.
    This op creates a set of anchor boxes by placing a "basis" collection of
    boxes with user-specified scales and aspect ratios centered at evenly
    distributed points along a grid.  The basis collection is specified via the
    scale and aspect_ratios arguments.  For example, setting scales=[.1, .2, .2]
    and aspect ratios = [2,2,1/2] means that we create three boxes: one with scale
    .1, aspect ratio 2, one with scale .2, aspect ratio 2, and one with scale .2
    and aspect ratio 1/2.  Each box is multiplied by "base_anchor_size" before
    placing it over its respective center.
    Grid points are specified via grid_height, grid_width parameters as well as
    the anchor_stride and anchor_offset parameters.
    Args:
      grid_height: size of the grid in the y direction (int or int scalar tensor)
      grid_width: size of the grid in the x direction (int or int scalar tensor)
      scales: a 1-d  (float) tensor representing the scale of each box in the
        basis set.
      aspect_ratios: a 1-d (float) tensor representing the aspect ratio of each
        box in the basis set.  The length of the scales and aspect_ratios tensors
        must be equal.
      base_anchor_size: base anchor size as [height, width]
        (float tensor of shape [2])
      anchor_stride: difference in centers between base anchors for adjacent grid
                     positions (float tensor of shape [2])
      anchor_offset: center of the anchor with scale and aspect ratio 1 for the
                     upper left element of the grid, this should be zero for
                     feature networks with only VALID padding and even receptive
                     field size, but may need some additional calculation if other
                     padding is used (float tensor of shape [2])
    Returns:
      a BoxList holding a collection of N anchor boxes
    """
    aspect_ratios = torch.as_tensor(aspect_ratios, dtype=torch.float32)
    scales = torch.as_tensor(scales, dtype=torch.float32)

    ratio_sqrts = torch.sqrt(aspect_ratios)
    heights = scales / ratio_sqrts * base_anchor_size[0]
    widths = scales * ratio_sqrts * base_anchor_size[1]

    # Get a grid of box centers
    y_centers = torch.arange(grid_height, dtype=torch.float32)
    y_centers = y_centers * anchor_stride[0] + anchor_offset[0]
    x_centers = torch.arange(grid_width, dtype=torch.float32)
    x_centers = x_centers * anchor_stride[1] + anchor_offset[1]

    x_centers, y_centers = meshgrid(x_centers, y_centers)

    widths_grid, x_centers_grid = meshgrid(widths, x_centers)
    heights_grid, y_centers_grid = meshgrid(heights, y_centers)

    bbox_centers = torch.stack([y_centers_grid, x_centers_grid], dim=3)
    bbox_sizes = torch.stack([heights_grid, widths_grid], dim=3)
    bbox_centers = torch.reshape(bbox_centers, [-1, 2])
    bbox_sizes = torch.reshape(bbox_sizes, [-1, 2])
    bbox_corners = _center_size_bbox_to_corners_bbox(bbox_centers, bbox_sizes)
    return bbox_corners


def _center_size_bbox_to_corners_bbox(centers, sizes):
    """Converts bbox center-size representation to corners representation.
    Args:
      centers: a tensor with shape [N, 2] representing bounding box centers
      sizes: a tensor with shape [N, 2] representing bounding boxes
    Returns:
      corners: tensor with shape [N, 4] representing bounding boxes in corners
        representation
    """
    return torch.cat([centers - 0.5 * sizes, centers + 0.5 * sizes], 1)


def create_ssd_anchors(
    num_layers=6,
    min_scale=0.2,
    max_scale=0.95,
    scales=None,
    aspect_ratios=(1.0, 2.0, 1.0 / 2, 3.0, 1.0 / 3),
    interpolated_scale_aspect_ratio=1.0,
    base_anchor_size=None,
    anchor_strides=None,
    anchor_offsets=None,
    reduce_boxes_in_lowest_layer=True,
):
    """Creates MultipleGridAnchorGenerator for SSD anchors.
    This function instantiates a MultipleGridAnchorGenerator that reproduces
    ``default box`` construction proposed by Liu et al in the SSD paper.
    See Section 2.2 for details. Grid sizes are assumed to be passed in
    at generation time from finest resolution to coarsest resolution --- this is
    used to (linearly) interpolate scales of anchor boxes corresponding to the
    intermediate grid sizes.
    Anchors that are returned by calling the `generate` method on the returned
    MultipleGridAnchorGenerator object are always in normalized coordinates
    and clipped to the unit square: (i.e. all coordinates lie in [0, 1]x[0, 1]).
    Args:
      num_layers: integer number of grid layers to create anchors for (actual
        grid sizes passed in at generation time)
      min_scale: scale of anchors corresponding to finest resolution (float)
      max_scale: scale of anchors corresponding to coarsest resolution (float)
      scales: As list of anchor scales to use. When not None and not empty,
        min_scale and max_scale are not used.
      aspect_ratios: list or tuple of (float) aspect ratios to place on each
        grid point.
      interpolated_scale_aspect_ratio: An additional anchor is added with this
        aspect ratio and a scale interpolated between the scale for a layer
        and the scale for the next layer (1.0 for the last layer).
        This anchor is not included if this value is 0.
      base_anchor_size: base anchor size as [height, width].
        The height and width values are normalized to the minimum dimension of the
        input height and width, so that when the base anchor height equals the
        base anchor width, the resulting anchor is square even if the input image
        is not square.
      anchor_strides: list of pairs of strides in pixels (in y and x directions
        respectively). For example, setting anchor_strides=[(25, 25), (50, 50)]
        means that we want the anchors corresponding to the first layer to be
        strided by 25 pixels and those in the second layer to be strided by 50
        pixels in both y and x directions. If anchor_strides=None, they are set to
        be the reciprocal of the corresponding feature map shapes.
      anchor_offsets: list of pairs of offsets in pixels (in y and x directions
        respectively). The offset specifies where we want the center of the
        (0, 0)-th anchor to lie for each layer. For example, setting
        anchor_offsets=[(10, 10), (20, 20)]) means that we want the
        (0, 0)-th anchor of the first layer to lie at (10, 10) in pixel space
        and likewise that we want the (0, 0)-th anchor of the second layer to lie
        at (25, 25) in pixel space. If anchor_offsets=None, then they are set to
        be half of the corresponding anchor stride.
      reduce_boxes_in_lowest_layer: a boolean to indicate whether the fixed 3
        boxes per location is used in the lowest layer.
    Returns:
      a MultipleGridAnchorGenerator
    """
    if base_anchor_size is None:
        base_anchor_size = [1.0, 1.0]
    base_anchor_size = torch.tensor(base_anchor_size, dtype=torch.float32)
    box_specs_list = []
    if scales is None or not scales:
        scales = [
            min_scale + (max_scale - min_scale) * i / (num_layers - 1) for i in range(num_layers)
        ] + [1.0]
    else:
        # Add 1.0 to the end, which will only be used in scale_next below and used
        # for computing an interpolated scale for the largest scale in the list.
        scales += [1.0]

    for layer, scale, scale_next in zip(range(num_layers), scales[:-1], scales[1:]):
        layer_box_specs = []
        if layer == 0 and reduce_boxes_in_lowest_layer:
            layer_box_specs = [(0.1, 1.0), (scale, 2.0), (scale, 0.5)]
        else:
            for aspect_ratio in aspect_ratios:
                layer_box_specs.append((scale, aspect_ratio))
            # Add one more anchor, with a scale between the current scale, and the
            # scale for the next layer, with a specified aspect ratio (1.0 by
            # default).
            if interpolated_scale_aspect_ratio > 0.0:
                layer_box_specs.append(
                    (np.sqrt(scale * scale_next), interpolated_scale_aspect_ratio)
                )
        box_specs_list.append(layer_box_specs)

    return MultipleGridAnchorGenerator(
        box_specs_list, base_anchor_size, anchor_strides, anchor_offsets
    )


class MultipleGridAnchorGenerator(object):
    """Generate a grid of anchors for multiple CNN layers."""

    def __init__(
        self,
        box_specs_list,
        base_anchor_size=None,
        anchor_strides=None,
        anchor_offsets=None,
        clip_window=None,
    ):
        """Constructs a MultipleGridAnchorGenerator.
        To construct anchors, at multiple grid resolutions, one must provide a
        list of feature_map_shape_list (e.g., [(8, 8), (4, 4)]), and for each grid
        size, a corresponding list of (scale, aspect ratio) box specifications.
        For example:
        box_specs_list = [[(.1, 1.0), (.1, 2.0)],  # for 8x8 grid
                          [(.2, 1.0), (.3, 1.0), (.2, 2.0)]]  # for 4x4 grid
        To support the fully convolutional setting, we pass grid sizes in at
        generation time, while scale and aspect ratios are fixed at construction
        time.
        Args:
          box_specs_list: list of list of (scale, aspect ratio) pairs with the
            outside list having the same number of entries as feature_map_shape_list
            (which is passed in at generation time).
          base_anchor_size: base anchor size as [height, width]
                            (length-2 float tensor, default=[1.0, 1.0]).
                            The height and width values are normalized to the
                            minimum dimension of the input height and width, so that
                            when the base anchor height equals the base anchor
                            width, the resulting anchor is square even if the input
                            image is not square.
          anchor_strides: list of pairs of strides in pixels (in y and x directions
            respectively). For example, setting anchor_strides=[(25, 25), (50, 50)]
            means that we want the anchors corresponding to the first layer to be
            strided by 25 pixels and those in the second layer to be strided by 50
            pixels in both y and x directions. If anchor_strides=None, they are set
            to be the reciprocal of the corresponding feature map shapes.
          anchor_offsets: list of pairs of offsets in pixels (in y and x directions
            respectively). The offset specifies where we want the center of the
            (0, 0)-th anchor to lie for each layer. For example, setting
            anchor_offsets=[(10, 10), (20, 20)]) means that we want the
            (0, 0)-th anchor of the first layer to lie at (10, 10) in pixel space
            and likewise that we want the (0, 0)-th anchor of the second layer to
            lie at (25, 25) in pixel space. If anchor_offsets=None, then they are
            set to be half of the corresponding anchor stride.
          clip_window: a tensor of shape [4] specifying a window to which all
            anchors should be clipped. If clip_window is None, then no clipping
            is performed.
        Raises:
          ValueError: if box_specs_list is not a list of list of pairs
          ValueError: if clip_window is not either None or a tensor of shape [4]
        """
        if isinstance(box_specs_list, list) and all(
            [isinstance(list_item, list) for list_item in box_specs_list]
        ):
            self._box_specs = box_specs_list
        else:
            raise ValueError("box_specs_list is expected to be a " "list of lists of pairs")
        if base_anchor_size is None:
            base_anchor_size = torch.tensor([256, 256], dtype=torch.float32)
        self._base_anchor_size = base_anchor_size
        self._anchor_strides = anchor_strides
        self._anchor_offsets = anchor_offsets
        if clip_window is not None and list(clip_window.shape) != [4]:
            raise ValueError("clip_window must either be None or a shape [4] tensor")
        self._clip_window = clip_window
        self._scales = []
        self._aspect_ratios = []
        for box_spec in self._box_specs:
            if not all([isinstance(entry, tuple) and len(entry) == 2 for entry in box_spec]):
                raise ValueError("box_specs_list is expected to be a " "list of lists of pairs")
            scales, aspect_ratios = zip(*box_spec)
            self._scales.append(scales)
            self._aspect_ratios.append(aspect_ratios)

        for arg, arg_name in zip(
            [self._anchor_strides, self._anchor_offsets],
            ["anchor_strides", "anchor_offsets"],
        ):
            if arg and not (isinstance(arg, list) and len(arg) == len(self._box_specs)):
                raise ValueError(
                    "%s must be a list with the same length " "as self._box_specs" % arg_name
                )
            if arg and not all(
                [isinstance(list_item, tuple) and len(list_item) == 2 for list_item in arg]
            ):
                raise ValueError("%s must be a list of pairs." % arg_name)

    def _generate(self, feature_map_shape_list, im_height=1, im_width=1):
        """Generates a collection of bounding boxes to be used as anchors.
        The number of anchors generated for a single grid with shape MxM where we
        place k boxes over each grid center is k*M^2 and thus the total number of
        anchors is the sum over all grids. In our box_specs_list example
        (see the constructor docstring), we would place two boxes over each grid
        point on an 8x8 grid and three boxes over each grid point on a 4x4 grid and
        thus end up with 2*8^2 + 3*4^2 = 176 anchors in total. The layout of the
        output anchors follows the order of how the grid sizes and box_specs are
        specified (with box_spec index varying the fastest, followed by width
        index, then height index, then grid index).
        Args:
          feature_map_shape_list: list of pairs of convnet layer resolutions in the
            format [(height_0, width_0), (height_1, width_1), ...]. For example,
            setting feature_map_shape_list=[(8, 8), (7, 7)] asks for anchors that
            correspond to an 8x8 layer followed by a 7x7 layer.
          im_height: the height of the image to generate the grid for. If both
            im_height and im_width are 1, the generated anchors default to
            absolute coordinates, otherwise normalized coordinates are produced.
          im_width: the width of the image to generate the grid for. If both
            im_height and im_width are 1, the generated anchors default to
            absolute coordinates, otherwise normalized coordinates are produced.
        Returns:
          boxes_list: a list of BoxLists each holding anchor boxes corresponding to
            the input feature map shapes.
        Raises:
          ValueError: if feature_map_shape_list, box_specs_list do not have the same
            length.
          ValueError: if feature_map_shape_list does not consist of pairs of
            integers
        """
        if not (
            isinstance(feature_map_shape_list, list)
            and len(feature_map_shape_list) == len(self._box_specs)
        ):
            raise ValueError(
                "feature_map_shape_list must be a list with the same " "length as self._box_specs"
            )
        if not all(
            [
                isinstance(list_item, tuple) and len(list_item) == 2
                for list_item in feature_map_shape_list
            ]
        ):
            raise ValueError("feature_map_shape_list must be a list of pairs.")

        im_height = float(im_height)
        im_width = float(im_width)

        if not self._anchor_strides:
            anchor_strides = [
                (1.0 / float(pair[0]), 1.0 / float(pair[1])) for pair in feature_map_shape_list
            ]
        else:
            anchor_strides = [
                (float(stride[0]) / im_height, float(stride[1]) / im_width)
                for stride in self._anchor_strides
            ]
        if not self._anchor_offsets:
            anchor_offsets = [(0.5 * stride[0], 0.5 * stride[1]) for stride in anchor_strides]
        else:
            anchor_offsets = [
                (float(offset[0]) / im_height, float(offset[1]) / im_width)
                for offset in self._anchor_offsets
            ]

        for arg, arg_name in zip(
            [anchor_strides, anchor_offsets], ["anchor_strides", "anchor_offsets"]
        ):
            if not (isinstance(arg, list) and len(arg) == len(self._box_specs)):
                raise ValueError(
                    "%s must be a list with the same length " "as self._box_specs" % arg_name
                )
            if not all([isinstance(list_item, tuple) and len(list_item) == 2 for list_item in arg]):
                raise ValueError("%s must be a list of pairs." % arg_name)

        anchor_grid_list = []
        min_im_shape = min(im_height, im_width)
        scale_height = min_im_shape / im_height
        scale_width = min_im_shape / im_width
        base_anchor_size = [
            scale_height * self._base_anchor_size[0],
            scale_width * self._base_anchor_size[1],
        ]
        for feature_map_index, (grid_size, scales, aspect_ratios, stride, offset) in enumerate(
            zip(
                feature_map_shape_list,
                self._scales,
                self._aspect_ratios,
                anchor_strides,
                anchor_offsets,
            )
        ):
            tiled_anchors = tile_anchors(
                grid_height=grid_size[0],
                grid_width=grid_size[1],
                scales=scales,
                aspect_ratios=aspect_ratios,
                base_anchor_size=base_anchor_size,
                anchor_stride=stride,
                anchor_offset=offset,
            )
            if self._clip_window is not None:
                raise NotImplementedError("Oups!")
            num_anchors_in_layer = len(tiled_anchors)
            anchorndices = feature_map_index * torch.ones(num_anchors_in_layer)  # noqa: F841
            anchor_grid_list.append(tiled_anchors)

        return anchor_grid_list



from typing import Any, ClassVar, Dict, List, Sequence, Tuple, Type, Union

import cv2
import numpy as np

from ..._utils import validate_postprocessor_type
from ...types import (
    Format,
    Metadata,
    Platform,
    PoseEstimationModel,
    PostProcessor,
    PreProcessor,
    Publication,
)
from ..preprocess import read_image_opencv_if_needed
from .postprocess import PoseEstimationResult, YOLOv7w6PosePostProcessor

_INPUT_SIZE = (640, 384)


def _letterbox(
    im: np.ndarray,
    new_shape: Tuple[int, int] = _INPUT_SIZE,
    color: Tuple[int, int, int] = (114, 114, 114),
    auto: bool = True,
    scaleFill: bool = False,
    scaleup: bool = True,
    stride: int = 32,
) -> Tuple[np.ndarray, Tuple[float, float], Tuple[int, int]]:
    """Letterboxing is a method for fitting an image within
       a constrained space dictated by the model requirements.
       If some axis is smaller than new_shape, it will be filled with color.
    Args:
        im (np.ndarray): a numpy image. Its shape must be Channel x Height x Width.
        new_shape (Tuple[int, int], optional): Targeted Image size. Defaults to default input size
            (640, 640).
        color (Tuple[int, int, int], optional): Padding Color Value. Defaults to (114, 114, 114).
        auto (bool, optional): If True, calculate padding width and height along to stride. Default
            to True.
        scaleFill (bool, optional): If True and auto is False, stretch an give image to target shape
            without any padding. Default to False.
        scaleup (bool, optional): If True, only scale down. Default to True.
        stride (int, optional): an output size of an image must be divied by a stride it is
            dependent on a model. This is only valid for auto is True. Default to 32.

    Returns:
        Tuple[np.ndarray, Tuple[float, float], Tuple[int, int]]:
            The first element is an padded-resized image. The second element is a resized image.
            The last element is padded sizes, respectively width and height.
    """
    # Resize and pad image while meeting stride-multiple constraints
    shape = im.shape[:2]  # current shape [height, width]
    if isinstance(new_shape, int):
        new_shape = (new_shape, new_shape)

    # Scale ratio (new / old)
    r = min(new_shape[0] / shape[0], new_shape[1] / shape[1])
    if not scaleup:  # only scale down, do not scale up (for better val mAP)
        r = min(r, 1.0)

    # Compute padding
    ratio = r, r  # width, height ratios
    new_unpad = int(round(shape[1] * r)), int(round(shape[0] * r))
    dw, dh = new_shape[1] - new_unpad[0], new_shape[0] - new_unpad[1]  # wh padding
    if auto:  # minimum rectangle
        dw, dh = np.mod(dw, stride), np.mod(dh, stride)  # wh padding
    elif scaleFill:  # stretch
        dw, dh = 0.0, 0.0
        new_unpad = (new_shape[1], new_shape[0])
        ratio = new_shape[1] / shape[1], new_shape[0] / shape[0]  # width, height ratios

    # divide padding into 2 sides
    dw /= 2
    dh /= 2

    if shape[::-1] != new_unpad:  # resize
        im = cv2.resize(im, new_unpad, interpolation=cv2.INTER_LINEAR)
    top, bottom = int(round(dh - 0.1)), int(round(dh + 0.1))
    left, right = int(round(dw - 0.1)), int(round(dw + 0.1))
    im = cv2.copyMakeBorder(
        im, top, bottom, left, right, cv2.BORDER_CONSTANT, value=color
    )  # add border
    return im, ratio, (dw, dh)


def _resize(img, model_input_size):
    w, h = model_input_size
    return _letterbox(img, (h, w), auto=False)


class YOLOv7w6PosePreProcessor(PreProcessor):
    @staticmethod
    def __call__(
        images: Sequence[Union[str, np.ndarray]], with_scaling: bool = False
    ) -> Tuple[np.ndarray, List[Dict[str, Any]]]:
        """Preprocess input images to a batch of input tensors

        Args:
            images: Color images have (NHWC: Batch, Height, Width, Channel) dimensions.
            with_scaling: Whether to apply model-specific techniques that involve scaling the
                model's input and converting its data type to float32. Refer to the code to gain a
                precise understanding of the techniques used. Defaults to False.

        Returns:
            a pre-processed image, scales and padded sizes(width,height) per images.
                The first element is a stacked numpy array containing a batch of images.
                To learn more about the outputs of preprocess (i.e., model inputs),
                please refer to [YOLOv7w6Pose Inputs](yolov7_w6_pose.md#inputs).

                The second element is a list of dict objects about the original images.
                Each dict object has the following keys. 'scale' key of the returned dict has a
                rescaled ratio per width(=target/width) and height(=target/height), and the 'pad'
                key has padded width and height pixels. Specially, the last dictionary element of
                returning tuple will be passed to postprocessing as a parameter to calculate
                predicted coordinates on normalized coordinates back to an input image coordinator.
        """
        # image format must be chw
        batched_image = []
        batched_proc_params = []
        if isinstance(images, str):
            images = [images]
        for image in images:
            image = read_image_opencv_if_needed(image)
            assert image.dtype == np.uint8
            if with_scaling:
                image = image.astype(np.float32)
            image, (sx, sy), (padw, padh) = _resize(image, _INPUT_SIZE)
            image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)

            if with_scaling:
                image /= 255.0
            image = image.transpose([2, 0, 1])  # NHWC -> NCHW
            scale = sx
            batched_image.append(image)
            batched_proc_params.append({"scale": scale, "pad": (padw, padh)})

        return np.stack(batched_image, axis=0), batched_proc_params


class YOLOv7w6Pose(PoseEstimationModel):
    """YOLOv7 w6 Pose Estimation model"""

    postprocessor_map: ClassVar[Dict[Platform, Type[PostProcessor]]] = {
        Platform.PYTHON: YOLOv7w6PosePostProcessor,
    }

    @staticmethod
    def visualize(image: np.ndarray, results: List[PoseEstimationResult]):
        """This visualize function is an example of how to visualize the output of the model.
        It draws a skeleton of the human body on the input image in in-place manner.

        Args:
            image: an input image
            results: a list of PoseEstimationResult objects
        """

        keypoints = [
            "nose",
            "left_eye",
            "right_eye",
            "left_ear",
            "right_ear",
            "left_shoulder",
            "right_shoulder",
            "left_elbow",
            "right_elbow",
            "left_wrist",
            "right_wrist",
            "left_hip",
            "right_hip",
            "left_knee",
            "right_knee",
            "left_ankle",
            "right_ankle",
        ]
        palette = np.array(
            [
                [255, 128, 0],
                [255, 153, 51],
                [255, 178, 102],
                [230, 230, 0],
                [255, 153, 255],
                [153, 204, 255],
                [255, 102, 255],
                [255, 51, 255],
                [102, 178, 255],
                [51, 153, 255],
                [255, 153, 153],
                [255, 102, 102],
                [255, 51, 51],
                [153, 255, 153],
                [102, 255, 102],
                [51, 255, 51],
                [0, 255, 0],
                [0, 0, 255],
                [255, 0, 0],
                [255, 255, 255],
            ],
            np.int32,
        )

        skeletons = [
            [16, 14],
            [14, 12],
            [17, 15],
            [15, 13],
            [12, 13],
            [6, 12],
            [7, 13],
            [6, 7],
            [6, 8],
            [7, 9],
            [8, 10],
            [9, 11],
            [2, 3],
            [1, 2],
            [1, 3],
            [2, 4],
            [3, 5],
            [4, 6],
            [5, 7],
        ]

        pose_limb_color = palette[[9, 9, 9, 9, 7, 7, 7, 0, 0, 0, 0, 0, 16, 16, 16, 16, 16, 16, 16]]
        pose_kpt_color = palette[[16, 16, 16, 16, 16, 0, 0, 0, 0, 0, 0, 9, 9, 9, 9, 9, 9]]

        def to_int_position(keypoint):
            return tuple(map(int, [keypoint.x, keypoint.y]))

        def is_valid_keypoint(keypoint):
            return (
                keypoint.x % 640 != 0
                and keypoint.y % 640 != 0
                and keypoint.x >= 0
                and keypoint.y >= 0
            )

        for result in results:
            for color, keypoint_name in zip(pose_kpt_color, keypoints):
                point = getattr(result, keypoint_name)
                if is_valid_keypoint(point):
                    cv2.circle(
                        image,
                        to_int_position(point),
                        radius=3,
                        color=color.tolist(),
                        thickness=-1,
                    )

            for color, skeleton in zip(pose_limb_color, skeletons):
                pos1 = getattr(result, keypoints[skeleton[0] - 1])
                pos2 = getattr(result, keypoints[skeleton[1] - 1])
                if is_valid_keypoint(pos1) and is_valid_keypoint(pos2):
                    cv2.line(
                        image,
                        to_int_position(pos1),
                        to_int_position(pos2),
                        color.tolist(),
                        thickness=2,
                    )

    def __init__(self, *, postprocessor_type: Union[str, Platform] = Platform.PYTHON):
        postprocessor_type = Platform(postprocessor_type)
        validate_postprocessor_type(postprocessor_type, self.postprocessor_map.keys())
        super().__init__(
            name="YOLOv7w6Pose",
            format=Format.ONNX,
            family="YOLO",
            version="v7",
            metadata=Metadata(
                description="YOLOv7 w6 Pose Estimation model",
                publication=Publication(url="https://github.com/WongKinYiu/yolov7#pose-estimation"),
            ),
            preprocessor=YOLOv7w6PosePreProcessor(),
            postprocessor=self.postprocessor_map[postprocessor_type](),
        )
        self._artifact_name = "yolov7_w6_pose"



from abc import ABC, abstractmethod
import datetime
from enum import Enum
from functools import cached_property
from typing import Any, Dict, List, Literal, Optional, Sequence, Tuple

import numpy.typing as npt
from pydantic import BaseModel, ConfigDict, Field, computed_field, field_serializer
from typing_extensions import TypeAlias
import yaml

from ._utils import EXT_CALIB_YAML, EXT_ONNX, resolve_model_source, resolve_source

# Context type alias
Context: TypeAlias = Any


class Platform(str, Enum):
    """Implemented platform"""

    PYTHON = "PYTHON"
    C = "C"
    CPP = "CPP"
    RUST = "RUST"

    @classmethod
    def _missing_(cls, value):
        for member in cls:
            if member.value == value.upper():
                return member


class ModelTaskType(str, Enum):
    """Model's task type"""

    OBJECT_DETECTION = "OBJECT_DETECTION"
    IMAGE_CLASSIFICATION = "IMAGE_CLASSIFICATION"
    POSE_ESTIMATION = "POSE_ESTIMATION"


class Format(str, Enum):
    """Model binary format to represent the binary specified."""

    ONNX = "ONNX"
    TFLite = "TFLITE"


class PreProcessor(ABC):
    @abstractmethod
    def __call__(self, inputs: Any) -> Tuple[Sequence[npt.ArrayLike], Sequence[Context]]:
        ...


class PostProcessor(ABC):
    @abstractmethod
    def __call__(self, model_outputs: Sequence[npt.ArrayLike], contexts: Sequence[Context]) -> Any:
        ...


class RustPostProcessor(PostProcessor):
    platform = Platform.RUST


class PythonPostProcessor(PostProcessor):
    platform = Platform.PYTHON


class Publication(BaseModel, extra='forbid'):
    """Model publication information."""

    authors: Optional[List[str]] = None
    title: Optional[str] = None
    publisher: Optional[str] = None
    date: Optional[datetime.date] = None
    url: Optional[str] = None


class Metadata(BaseModel, extra='forbid'):
    """Model metadata to understand a model."""

    description: Optional[str] = None
    publication: Optional[Publication] = None


class Tags(BaseModel, extra='forbid'):
    """Model tags to understand a model."""

    content_type: Optional[str] = None


class Model(ABC, BaseModel):
    """Represent the artifacts and metadata of a neural network model

    Attributes:
        name: a name of this model
        task_type: the task type of this model
        format: the binary format type of model origin; e.g., ONNX, tflite
        family: the model family
        version: the model version
        metadata: the model metadata
        tags: the model tags
        origin: an origin f32 binary in ONNX or tflite. It can be used for compiling this model
            with or without quantization and proper compiler configuration
        tensor_name_to_range: the calibration ranges of each tensor in origin
        preprocessor: a preprocessor to preprocess input tensors
        postprocessor: a postprocessor to postprocess output tensors
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    name: str
    task_type: ModelTaskType
    format: Format
    family: Optional[str] = None
    version: Optional[str] = None
    metadata: Optional[Metadata] = None
    tags: Optional[Tags] = None

    _artifact_name: str

    preprocessor: PreProcessor = Field(..., repr=False, exclude=True)
    postprocessor: PostProcessor = Field(..., repr=False, exclude=True)

    def preprocess(self, *args, **kwargs) -> Tuple[Sequence[npt.ArrayLike], Sequence[Context]]:
        """Preprocess input tensors. Input of this function varies from model to model

        Returns:
            A tuple that contains list of preprocessed input tensors and contexts"""
        return self.preprocessor(*args, **kwargs)

    def postprocess(self, *args, **kwargs):
        """Postprocess output tensors"""
        return self.postprocessor(*args, **kwargs)

    @computed_field(repr=False)
    @cached_property
    def origin(self) -> bytes:
        return resolve_source(self._artifact_name, EXT_ONNX)

    @computed_field(repr=False)
    @cached_property
    def tensor_name_to_range(self) -> Dict[str, List[float]]:
        calib_yaml = resolve_source(self._artifact_name, EXT_CALIB_YAML)
        return yaml.full_load(calib_yaml)

    def model_source(self, num_pe: Literal[1, 2] = 2) -> bytes:
        """Returns an executable binary for furiosa runtime and NPU. It can be
            directly fed to `furiosa.runtime.create_runner`. If model binary is not compiled yet,
            it will be quantized & compiled automatically if possible

        Args:
            num_pe: number of PE to be used.
        """
        if num_pe not in (1, 2):
            raise ValueError(f"Invalid num_pe: {num_pe}")

        # TODO: Add in-memory cached value(like cached_property), currently uses disk-cached value
        return resolve_model_source(self._artifact_name, num_pe=num_pe)

    def resolve_all(self):
        """Resolve all non-cached properties(origin, tensor_name_to_range, model_sources)"""
        _ = self.origin, self.tensor_name_to_range
        for num_pe in (1, 2):
            _ = self.model_source(num_pe=num_pe)

    @field_serializer('format')
    def serialize_format(self, format: Format):
        return format.value

    @field_serializer('task_type')
    def serialize_task_type(self, task_type: ModelTaskType):
        return task_type.value


class ObjectDetectionModel(Model, ABC):
    """Object Detection Model Base Class"""

    def __init__(self, *args, **kwargs):
        super().__init__(task_type=ModelTaskType.OBJECT_DETECTION, *args, **kwargs)


class ImageClassificationModel(Model, ABC):
    """Image Classification Model Base Class"""

    def __init__(self, *args, **kwargs):
        super().__init__(task_type=ModelTaskType.IMAGE_CLASSIFICATION, *args, **kwargs)


class PoseEstimationModel(Model, ABC):
    """Pose Estimation Model Base Class"""

    def __init__(self, *args, **kwargs):
        super().__init__(task_type=ModelTaskType.POSE_ESTIMATION, *args, **kwargs)



from pathlib import Path
from typing import Callable, List, Optional, Type

from tabulate import tabulate
import typer
import yaml

from . import api
from .. import __version__ as models_version
from ..types import ImageClassificationModel, Model, ObjectDetectionModel, PoseEstimationModel

EXAMPLE: str = """Examples:\n\n\n
# List available models\n
`furiosa-models list`\n\n\n
# List Object Detection models\n
`furiosa-models list -t detect`\n\n\n
# Describe SSDResNet34 model\n
`furiosa-models desc SSDResNet34`\n\n\n
# Run SSDResNet34 for images in `./input` directory\n
`furiosa-models bench ssd-resnet34 ./input/`\n\n\n
# Run YOLOv7w6Pose REST API server for `0.0.0.0:8080`\n
`furiosa-models serve yolov7w6pose --host 0.0.0.0 --port 8080`
"""

app = typer.Typer(
    help=f"FuriosaAI Model Zoo CLI --- v{models_version}", epilog=EXAMPLE, add_completion=False
)


def get_model_list(table: List[List[str]]):
    header = ["Model name", "Model description", "Task type", "Available postprocesses"]
    return tabulate(table, headers=header, tablefmt="pretty")


def resolve_input_paths(input_path: Path) -> List[str]:
    """Create input file list"""
    if input_path.is_file():
        return [str(input_path)]
    elif input_path.is_dir():
        # Directory may contain image files
        image_extensions = {".jpg", ".jpeg", ".png"}
        return [
            str(p.resolve())
            for p in input_path.glob("**/*")
            if p.suffix.lower() in image_extensions
        ]
    else:
        typer.echo(f"Invalid input path '{str(input_path)}'")
        raise typer.Exit(1)


def get_filter(filter_type: Optional[str]) -> Callable[..., bool]:
    if filter_type is None:
        return lambda _: True
    elif "detect" in filter_type.lower():
        return lambda x: issubclass(x, ObjectDetectionModel)
    elif "classif" in filter_type.lower():
        return lambda x: issubclass(x, ImageClassificationModel)
    elif "pose" in filter_type.lower():
        return lambda x: issubclass(x, PoseEstimationModel)
    else:
        typer.echo(f"Unknown type filter '{filter_type}', showing all models...")
        return lambda _: True


def get_model_or_exit(model_name: str) -> Model:
    model = api.get_model(model_name)
    if model is None:
        typer.echo(f"Model name '{model_name}' not found")
        raise typer.Exit(1)
    return model


def describe_model(model_cls: Type[Model]) -> str:
    model = model_cls()
    include = {"name", "format", "family", "version", "metadata", "tags"}
    output = []
    output.append(yaml.dump(model.model_dump(include=include, exclude_none=True), sort_keys=False))
    output.append(f"task type: {api.prettified_task_type(model)}\n")
    available_postprocs = ', '.join(
        map(lambda x: x.name.capitalize(), model.postprocessor_map.keys())
    )
    output.append(f"available postprocess versions: {available_postprocs}")
    return ''.join(output)


@app.command("list", help="List available models")
def list_models(
    filter_type: Optional[str] = typer.Argument(
        None, help="Limits the task type (ex. classify, detect, pose)"
    )
):
    typer.echo(get_model_list(api.get_model_list(filter_func=get_filter(filter_type))))


@app.command("desc", help="Describe a model")
def describe_model_cmd(model_name: str):
    model_cls = get_model_or_exit(model_name)
    typer.echo(describe_model(model_cls))


@app.command("bench", help="Run benchmark on a model")
def benchmark_model(
    model: str,
    input_path: Path,
    postprocess: Optional[str] = typer.Option(
        None, "--postprocess", help="Specifies a postprocess implementation"
    ),
    device_str: Optional[str] = typer.Option(
        None, "--devices", help="Specifies devices to run the model (ex. 'warboy(2)*1')"
    ),
):
    input_paths = resolve_input_paths(Path(input_path))
    if len(input_paths) == 0:
        typer.echo(f"No input files found in '{input_path}'")
        raise typer.Exit(code=1)
    typer.echo(f"Collected input paths: {input_paths}")
    model_cls = get_model_or_exit(model)
    api.run_inferences(model_cls, input_paths, postprocess, device_str)


@app.command("serve", help="Open a REST API server for a model")
def serve_model(
    model: str,
    postprocess: Optional[str] = typer.Option(
        None, "--postprocess", help="Specifies a postprocess implementation"
    ),
    host: str = typer.Option("0.0.0.0", "--host", help="Specifies a host address"),
    port: int = typer.Option(8000, "--port", help="Specifies a port number"),
    device_str: Optional[str] = typer.Option(
        None, "--devices", help="Specifies devices to run the model (ex. 'warboy(2)*1')"
    ),
):
    model_cls = get_model_or_exit(model)
    api.serve_model(model_cls, postprocess=postprocess, host=host, port=port, device_str=device_str)


if __name__ == "__main__":
    app()



from pathlib import Path
from typing import Any, ClassVar, Dict, List, Sequence, Tuple, Type, Union

from PIL import Image, ImageOps
import numpy as np
import numpy.typing as npt

from ..._utils import validate_postprocessor_type
from ...types import (
    Format,
    ImageClassificationModel,
    Metadata,
    Platform,
    PostProcessor,
    PreProcessor,
    Publication,
    PythonPostProcessor,
)
from ..common.datasets import imagenet1k

IMAGENET_DEFAULT_MEAN = np.array((0.485, 0.456, 0.406), dtype=np.float32)[:, np.newaxis, np.newaxis]
IMAGENET_DEFAULT_STD = np.array((0.229, 0.224, 0.225), dtype=np.float32)[:, np.newaxis, np.newaxis]

CLASSES: List[str] = imagenet1k.ImageNet1k_CLASSES
INPUT_SIZE = 384


def resize(image: Image.Image, size: int, interpolation: Image.Resampling) -> Image.Image:
    width, height = image.size
    short, long = (width, height) if width <= height else (height, width)
    requested_new_short = size

    new_short, new_long = requested_new_short, int(requested_new_short * long / short)
    new_width, new_height = (new_short, new_long) if width <= height else (new_long, new_short)
    return image.resize((new_width, new_height), interpolation)


def center_crop(image: Image.Image, output_size: int) -> Image.Image:
    image_width, image_height = image.size
    crop_height, crop_width = output_size, output_size

    if crop_width > image_width or crop_height > image_height:
        padding_ltrb = (
            (crop_width - image_width) // 2 if crop_width > image_width else 0,
            (crop_height - image_height) // 2 if crop_height > image_height else 0,
            (crop_width - image_width + 1) // 2 if crop_width > image_width else 0,
            (crop_height - image_height + 1) // 2 if crop_height > image_height else 0,
        )
        image = ImageOps.expand(image, padding_ltrb, fill=(0, 0, 0))
        image_width, image_height = image.size
        if crop_width == image_width and crop_height == image_height:
            return image

    crop_top = int(round((image_height - crop_height) / 2.0))
    crop_left = int(round((image_width - crop_width) / 2.0))
    return image.crop((crop_left, crop_top, crop_left + crop_width, crop_top + crop_height))


def normalize(image: Image.Image) -> np.ndarray:
    image -= IMAGENET_DEFAULT_MEAN
    image /= IMAGENET_DEFAULT_STD
    return image


class EfficientNetV2sPreProcessor(PreProcessor):
    @staticmethod
    def __call__(
        image: Union[str, Path, npt.ArrayLike], with_scaling: bool = False
    ) -> Tuple[np.ndarray, None]:
        """Read and preprocess an image located at image_path.

        Args:
            image: A path of an image.
            with_scaling: Whether to apply model-specific techniques that involve scaling the
                model's input and converting its data type to float32. Refer to the code to gain a
                precise understanding of the techniques used. Defaults to False.

        Returns:
            The first element of the tuple is a numpy array that meets the input requirements of the
                model. The second element of the tuple is unused in this model and has no value.
                To learn more information about the output numpy array, please refer to
                [Inputs](efficientnet_v2_s.md#inputs).

        """

        if isinstance(image, (str, Path)):
            image = Image.open(image).convert("RGB")

        image = resize(image, INPUT_SIZE, Image.Resampling.BILINEAR)
        image = center_crop(image, INPUT_SIZE)

        image = np.ascontiguousarray(image)
        assert image.dtype == np.uint8
        data = np.transpose(image, (2, 0, 1))

        if with_scaling:
            data = data.astype(np.float32) / 255
            data = normalize(data)

        return np.expand_dims(data, axis=0), None


class EfficientNetV2sPostProcessor(PythonPostProcessor):
    def __call__(self, model_outputs: Sequence[npt.ArrayLike], contexts: Any = None) -> str:
        """Convert the outputs of a model to a label string, such as car and cat.

        Args:
            model_outputs: the outputs of the model.
                Please learn more about the output of model,
                please refer to [Outputs](efficientnet_b0.md#outputs).

        Returns:
            str: A classified label, e.g., "tabby, tabby cat".
        """

        return CLASSES[int(np.argsort(model_outputs[0], axis=1)[:, ::-1][0, 0])]


class EfficientNetV2s(ImageClassificationModel):
    """EfficientNetV2-s model"""

    postprocessor_map: ClassVar[Dict[Platform, Type[PostProcessor]]] = {
        Platform.PYTHON: EfficientNetV2sPostProcessor,
    }

    def __init__(self, *, postprocessor_type: Union[str, Platform] = Platform.PYTHON):
        postprocessor_type = Platform(postprocessor_type)
        validate_postprocessor_type(postprocessor_type, self.postprocessor_map.keys())
        super().__init__(
            name="EfficientNetV2s",
            format=Format.ONNX,
            family="EfficientNetV2",
            version="s",
            metadata=Metadata(
                description="EfficientNetV2s ImageNet-1K",
                publication=Publication(url="https://arxiv.org/abs/2104.00298"),
            ),
            preprocessor=EfficientNetV2sPreProcessor(),
            postprocessor=self.postprocessor_map[postprocessor_type](),
        )

        self._artifact_name = "efficientnet_v2_s"
